package messif.objects.annotation;

import java.util.ArrayList;
import java.util.List;
import messif.objects.annotation.api.AnnotationProcessor;
import messif.objects.annotation.api.Annotator;
import messif.objects.classification.ClassificationException;
import messif.objects.classification.Classifier;
import messif.statistics.OperationStatistics;
import messif.statistics.StatisticTimer;
import messif.statistics.Statistics;
import messif.utility.ModifiableParametric;
import messif.utility.Parametric;

/**
 * Provides an annotation via executing a pipeline of several annotation classifiers.
 * The annotation initializer convert the annotation source object {@code T} to
 * the initial {@lin Annotation}. Then additional expanders, transformers, or reducers
 * can be added to the pipeline.
 *
 * @param <T> the class of the input instance for the annotation
 * @author Michal Batko, Masaryk University, Brno, Czech Republic, batko@fi.muni.cz
 */
public class AnnotationPipeline<T> implements Annotator<T, Annotation> {
    /** Annotator that provides the initial annotation that will be processed by the pipeline */
    private final Annotator<T, ?> initialAnnotator;
    /** Annotation processors that are applied one by one on the initial annotation */
    private final List<AnnotationProcessor<?, ?>> pipeline;
    /** List of names of the parameter into which the annotation result of that step is stored */
    private final List<String> parameterNames;

    /**
     * Creates a new instance of the pipeline with the given initial annotator.
     * @param initialAnnotator the annotator that provides the initial annotation that will be processed by the pipeline
     * @param parameterName the name of the parameter into which the annotation result of the initial annotator is stored
     */
    public AnnotationPipeline(Annotator<T, ?> initialAnnotator, String parameterName) {
        if (initialAnnotator == null)
            throw new NullPointerException();
        this.initialAnnotator = initialAnnotator;
        this.pipeline = new ArrayList<AnnotationProcessor<?, ?>>();
        this.parameterNames = new ArrayList<>();
        this.parameterNames.add(parameterName);
    }

    /**
     * Creates a new instance of the pipeline with the given initial annotator.
     * @param initialAnnotator the annotator that provides the initial annotation that will be processed by the pipeline
     */
    public AnnotationPipeline(Annotator<T, ?> initialAnnotator) {
        this(initialAnnotator, null);
    }

    /**
     * Adds a processor to this pipeline.
     * @param processor the processor to add
     * @param parameterName the name of the parameter into which the annotation result of the added processor is stored
     * @throws IllegalArgumentException if the processor input is not compatible with the current pipeline output
     */
    public void add(AnnotationProcessor<?, ?> processor, String parameterName) throws IllegalArgumentException {
        if (!processor.getAnnotationInputClass().isAssignableFrom(getAnnotationOutputClass()))
            throw new IllegalArgumentException("Processor " + processor.getClass().getSimpleName() + " cannot accept " + getAnnotationOutputClass() + " as input");
        this.pipeline.add(processor);
        this.parameterNames.add(parameterName);
    }

    /**
     * Adds a processor to this pipeline.
     * @param processor the processor to add
     * @throws IllegalArgumentException if the processor input is not compatible with the current pipeline output
     */
    public final void add(AnnotationProcessor<?, ?> processor) throws IllegalArgumentException {
        add(processor, null);
    }

    /**
     * Removes the last processor in this pipeline.
     * @return the removed processor
     */
    public AnnotationProcessor<?, ?> removeLast() {
        if (pipeline.isEmpty())
            return null;
        return pipeline.remove(pipeline.size() - 1);
    }

    /**
     * Returns the last annotator in the pipeline.
     * @return the last annotator in the pipeline
     */
    protected Annotator<?, ?> getLastAnnotator() {
        if (pipeline.isEmpty())
            return initialAnnotator;
        return pipeline.get(pipeline.size() - 1);
    }

    /**
     * Returns a {@link StatisticTimer} for the given classifier.
     * @param order the order of the classifier for which to get timer
     * @param start flag whether to start the timer immediately
     * @param classifier the classifier to get the timer for
     * @return timer instance
     */
    private StatisticTimer getStatisticTimer(int order, boolean start, Classifier<?, ?> classifier) {
        if (!Statistics.isEnabledGlobally())
            return null;
        StatisticTimer timer = OperationStatistics.getOpStatistics("AnnotationTimer." + order + "." + classifier.getClass().getSimpleName(), StatisticTimer.class);
        if (start)
            timer.start();
        return timer;
    }

    /**
     * Store the given annotation into the parameters using the name specified
     * in the respective item of {@link #parameterNames}.
     * @param processorIndex the processor index to use to get the name
     * @param value the value to store in the parameter
     * @param parameters the parameters where to store the annotation
     */
    private void setAnnotationParameter(int processorIndex, Object value, Parametric parameters) {
        if (!(parameters instanceof ModifiableParametric))
            return;
        if (processorIndex >= parameterNames.size())
            return;
        String parameterName = parameterNames.get(processorIndex);
        if (parameterName == null || parameterName.isEmpty())
            return;
        ((ModifiableParametric)parameters).setParameter(parameterName, value);
    }

    /**
     * Processes the input by the given annotator.
     * The statistics as well as parameters are updated if necessary.
     *
     * @param <I> the class of input annotation
     * @param <O> the class of output annotation
     * @param annotator the annotator to use on input
     * @param input the annotation to process
     * @param processorIndex the index of the annotator (used in statistics and parameter name)
     * @param parameters the annotation parameters
     * @return the processed annotation
     * @throws ClassificationException if there was an error during annotation
     */
    private <I, O extends Annotation> O annotate(Annotator<I, O> annotator, Object input, int processorIndex, Parametric parameters) throws ClassificationException {
        StatisticTimer timer = getStatisticTimer(processorIndex, true, annotator);
        O output = annotator.classify(annotator.getAnnotationInputClass().cast(input), parameters);
        if (timer != null)
            timer.stop();
        setAnnotationParameter(processorIndex, output, parameters);
        return output;
    }

    @Override
    public Annotation classify(T object, Parametric parameters) throws ClassificationException {
        int processorIndex = 0;
        Annotation result = annotate(initialAnnotator, object, processorIndex, parameters);
        for (Annotator<?, ?> annotator : pipeline) {
            processorIndex++;
            result = annotate(annotator, result, processorIndex, parameters);
        }
        return result;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }

    @Override
    public Class<? extends T> getAnnotationInputClass() {
        return initialAnnotator.getAnnotationInputClass();
    }

    @Override
    public Class<? extends Annotation> getAnnotationOutputClass() {
        return getLastAnnotator().getAnnotationOutputClass();
    }
}
