package messif.objects.annotation;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;
import java.util.Objects;

/**
 * Single term (a word or a word collocation) that is used to describe an object.
 *
 * @author Michal Batko <batko@fi.muni.cz>
 */
public class AnnotationTerm implements Comparable<AnnotationTerm> {
    private final String word;
    private final List<Object> originIdentifiers;
    private final boolean userSpecified;
    private final int originFrequency;

    public AnnotationTerm(String word, boolean userSpecified, int originFrequency, Collection<?> originIdentifiers) {
        if (word == null)
            throw new NullPointerException();
        this.word = word;
        this.userSpecified = userSpecified;
        this.originIdentifiers = originIdentifiers == null ? null : new ArrayList<>(originIdentifiers);
        this.originFrequency = originFrequency;
    }

    public AnnotationTerm(String word, Collection<?> originIdentifiers) {
        this(word, false, originIdentifiers.isEmpty() ? 1 : originIdentifiers.size(), originIdentifiers);
    }

    public AnnotationTerm(String word, boolean userSpecified, Object... originIdentifiers) {
        this(word, userSpecified, originIdentifiers.length == 0 ? 1 : originIdentifiers.length, Arrays.asList(originIdentifiers));
    }

    public AnnotationTerm(String word) {
        this(word, false);
    }

    public String getWord() {
        return word;
    }

    public int getOriginFrequency() {
        return originFrequency;
    }

    public int getOriginIdentifiersCount() {
        return originIdentifiers == null ? 0 : originIdentifiers.size();
    }

    public Object getOriginIdentifier(int index) {
        return originIdentifiers.get(index);
    }

    public Iterator<?> getOriginIterator() {
        return originIdentifiers == null ? Collections.emptyIterator() : originIdentifiers.iterator();
    }

    public List<Object> getOriginIdentifiers() {
        return originIdentifiers == null ? null : Collections.unmodifiableList(originIdentifiers);
    }

    public boolean isUserSpecified() {
        return userSpecified;
    }

    protected int superHashCode() {
        return super.hashCode();
    }

    protected boolean superEquals(Object obj) {
        return super.equals(obj);
    }

    @Override
    public int hashCode() {
        return word.hashCode();
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final AnnotationTerm other = (AnnotationTerm) obj;
        return Objects.equals(this.word, other.getWord());
    }

    /**
     * Single term (a word or a word collocation) that is used to describe an object.
     *
     * @author Michal Batko <batko@fi.muni.cz>
     */
    public static class AnnotationTermWithOrigin extends AnnotationTerm {

        public AnnotationTermWithOrigin(String word, boolean userSpecified, int originFrequency, Collection<?> originIdentifiers) {
            super(word, userSpecified, originFrequency, originIdentifiers);
        }

        public AnnotationTermWithOrigin(String word, Collection<?> originIdentifiers) {
            super(word, originIdentifiers);
        }

        public AnnotationTermWithOrigin(String word, boolean userSpecified, Object... originIdentifiers) {
            super(word, userSpecified, originIdentifiers);
        }

        @Override
        public int hashCode() {
            return superHashCode();
        }

        @Override
        public boolean equals(Object obj) {
            return superEquals(obj);
        }
    }

    @Override
    public int compareTo(AnnotationTerm o) {
        return word.compareTo(o.getWord());
    }

    /**
     * Returns an extended textual representation of this term.
     * The string consists of the word and the information about the origin frequency and identifiers.
     * @return an extended textual representation of this term
     */
    public String toStringExtended() {
        return word + "{" + originFrequency + ": " + originIdentifiers + "}";
    }

    @Override
    public String toString() {
        return word;
    }
}
