package messif.objects.annotation;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Set;
import messif.objects.LocalAbstractObject;
import net.didion.jwnl.data.Synset;

/**
 * The intermediate object for the Annotation Framework pipeline containing
 * AnnotationTerm to Synset mapping.
 *
 * @author Michal Batko <batko@fi.muni.cz>
 */
public class AnnotationWithSynsets extends Annotation {

    private static final long serialVersionUID = 1L;

    private final Map<AnnotationTerm, Set<Synset>> termSynsets;

    public AnnotationWithSynsets(LocalAbstractObject annotatedObject) {
        super(annotatedObject);
        this.termSynsets = new LinkedHashMap<>();
    }

    public AnnotationWithSynsets(Annotation originalAnnotation) {
        super(originalAnnotation);
        termSynsets = new LinkedHashMap<>();
    }

    public AnnotationWithSynsets(Annotation originalAnnotation, Map<AnnotationTerm, Set<Synset>> termSynsets) {
        super(originalAnnotation);
        this.termSynsets = termSynsets;
    }

    public Map<AnnotationTerm, Set<Synset>> getTermSynsets() {
        return termSynsets;
    }

    public int getTermSynsetCount(AnnotationTerm term) {
        Set<Synset> synsets = termSynsets.get(term);
        if (synsets == null) {
            return 0;
        }
        return synsets.size();
    }

    public Set<Synset> getSynsetsForTerm(AnnotationTerm term) {
        Set<Synset> synsets = termSynsets.get(term);
        if (synsets == null) {
            return Collections.emptySet();
        }
        return Collections.unmodifiableSet(synsets);
    }

    public Set<AnnotationTerm> getTermsForSynset(Synset synset) {
        Set<AnnotationTerm> ret = new HashSet<>();
        for (Map.Entry<AnnotationTerm, Set<Synset>> entry : termSynsets.entrySet()) {
            if (entry.getValue().contains(synset)) {
                ret.add(entry.getKey());
            }
        }
        return ret;
    }

    public AnnotationTerm getBestTermForSynset(Synset synset) throws NoSuchElementException {
        Iterator<AnnotationTerm> termIterator = getTermsForSynset(synset).iterator();
        if (!termIterator.hasNext())
            return null;
        AnnotationTerm bestTerm = termIterator.next();
        float bestConfidence = getConfidence(bestTerm);
        while (termIterator.hasNext()) {
            AnnotationTerm term = termIterator.next();
            float confidence = getConfidence(term);
            if (confidence > bestConfidence || (confidence == bestConfidence && term.getOriginFrequency() > bestTerm.getOriginFrequency())) {
                bestTerm = term;
                bestConfidence = confidence;
            }
        }
        return bestTerm;
    }

    public AnnotationTerm getMostFrequentTermForSynset(Synset synset) throws NoSuchElementException {
        Iterator<AnnotationTerm> termIterator = getTermsForSynset(synset).iterator();
        if (!termIterator.hasNext())
            return null;
        AnnotationTerm bestTerm = termIterator.next();
        while (termIterator.hasNext()) {
            AnnotationTerm term = termIterator.next();
            if (bestTerm.getOriginFrequency() < term.getOriginFrequency())
                bestTerm = term;
        }

        return bestTerm;
    }

    protected Set<Synset> getSynsetsCreate(AnnotationTerm term) {
        Set<Synset> synsets = termSynsets.get(term);
        if (synsets == null) {
            synsets = new LinkedHashSet<>();
            termSynsets.put(term, synsets);
        }
        return synsets;
    }

    public boolean addTermSynset(AnnotationTerm term, Synset synset) {
        return getSynsetsCreate(term).add(synset);
    }

    public boolean addTermSynsets(AnnotationTerm term, Collection<? extends Synset> synsets) {
        return getSynsetsCreate(term).addAll(synsets);
    }

}
