package messif.objects.annotation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import messif.objects.LocalAbstractObject;
import net.didion.jwnl.data.Synset;

/**
 * The intermediate and resulting object for the Annotation Framework pipeline.
 * Object contains a model of semantic network consisted of vertices (synsets
 * with probabilities) and edges (semantic links among vertices that distributed probability
 * values through the network).
 * @author Honza Botorek
 */
public class AnnotationWithSynsetsEdges extends AnnotationWithSynsets{
    private static final long serialVersionUID = 1L;

    private final Map<SynsetVertex, List<SynsetEdge>> semanticNetwork;

    public AnnotationWithSynsetsEdges(LocalAbstractObject annotatedObject) {
        super(annotatedObject);
        semanticNetwork = new LinkedHashMap<>();
    }

    public AnnotationWithSynsetsEdges(Annotation originalAnnotation, Map<AnnotationTerm, Set<Synset>> termSynsets) {
        super(originalAnnotation,termSynsets);
        semanticNetwork = new LinkedHashMap<>();
    }    

    public Map<SynsetVertex, List<SynsetEdge>> getSemanticNetwork() {
        return semanticNetwork;
    }

    public boolean addVertexToNetwork(SynsetVertex newSynset){
        if(networkContainsVertex(newSynset)) return false;
        else{
            List<SynsetEdge> edges = new ArrayList<>();
            semanticNetwork.put(newSynset, edges);
        }
        return true;
    }

    public boolean addSynsetEdgeToVertex(SynsetVertex syn, SynsetEdge edge){
        if(!networkContainsVertex(syn)) return false;
        if(edge == null) throw new NullPointerException("Edge can not be null");
        List<SynsetEdge> edges = semanticNetwork.get(syn);
        edges.add(edge);
        semanticNetwork.put(syn, edges);
        return true;
    }

    public boolean networkContainsVertex(SynsetVertex syn){
        return semanticNetwork.containsKey(syn);
    }

    public boolean copyVertex(SynsetVertex syn, AnnotationWithSynsetsEdges from, AnnotationWithSynsetsEdges to){
        if(to.networkContainsVertex(syn)) return false;
        to.addVertexToNetwork(syn);
        if(from.networkContainsVertex(syn)) {
            List<SynsetEdge> edges = from.getSemanticNetwork().get(syn);            
            for(SynsetEdge edge : edges) to.addSynsetEdgeToVertex(syn, edge);
        }
        return true;
    }

    @Override
    public String toString() {
        StringBuilder str = new StringBuilder("[");
        boolean isFirst = true;
        for (SynsetVertex vertex : SynsetVertex.sortByProbability(semanticNetwork.keySet())) {
            if (isFirst) {
                isFirst = false;
            } else {
                str.append(", ");
            }
            str.append(vertex.getSynsetWord()).append('(').append(vertex.getVertexProbability()).append(')');
        }
        return str.append(']').toString();
    }
}
