package messif.objects.annotation;

import messif.objects.annotation.api.Annotator;
import messif.objects.classification.ClassificationException;
import messif.objects.classification.Classifier;
import messif.utility.Parametric;

/**
 * Simple wrapper class that allows to convert a generic classifier that works with {@link AnnotationTerm}
 * to {@link Annotator}.
 *
 * @param <T> the type of object that this annotator accepts as input
 * @param <O> the type of annotation that this annotator provides as output
 * @author Michal Batko <batko@fi.muni.cz>
 */
public class ClassifierToAnnotator<T, O extends Annotation> implements Annotator<T, O> {
    /** Wrapped classifier that provides the annotation */
    private final Classifier<? super T, AnnotationTerm> classifier;
    /** Wrapped classifier output class, i.e. the {@link Annotation} */
    private final Class<? extends O> outputClass;

    /**
     * Factory method for creating the wrapper classifier.
     * @param <T> the type of object that the annotator accepts as input
     * @param <O> the type of annotation that the annotator provides as output
     * @param classifier the classifier to wrap
     * @param outputClass the classifier output class (must be a subclass of {@link Annotation}
     * @return the wrapped classifier
     */
    public static <T, O extends Annotation> Annotator<T, O> create(Classifier<T, AnnotationTerm> classifier, Class<? extends O> outputClass) {
        if (classifier instanceof Annotator) {
            @SuppressWarnings("unchecked")
            Annotator<T, O> annotator = (Annotator<T, O>)classifier; // This cast IS checked on the next line
            if (!outputClass.isAssignableFrom(annotator.getAnnotationOutputClass()))
                throw new ClassCastException("Annotator " + annotator + " does not return " + outputClass);
            return annotator;
        }
        return new ClassifierToAnnotator<>(classifier, outputClass);
    }

    /**
     * Creates a new wrapper for the given classifier.
     * @param classifier the classifier to wrap
     * @param outputClass the classifier output class (must be a subclass of {@link Annotation}
     */
    public ClassifierToAnnotator(Classifier<? super T, AnnotationTerm> classifier, Class<? extends O> outputClass) {
        this.classifier = classifier;
        this.outputClass = outputClass;
    }

    @Override
    public O classify(T object, Parametric parameters) throws ClassificationException {
        return outputClass.cast(classifier.classify(object, parameters));
    }

    @Override
    public Class<? extends T> getAnnotationInputClass() {
        return (Class)Object.class;
    }

    @Override
    public Class<? extends O> getAnnotationOutputClass() {
        return outputClass;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return classifier.getCategoriesClass();
    }

}
