package messif.objects.annotation;
    
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;
import java.util.Objects;
import net.didion.jwnl.data.Synset;
import net.didion.jwnl.data.Word;

/**
 * Component of the semantic graph.
 * Simple extension of original Synset object. Furthermore, it keeps a probability
 * value and information whether or not the synset was marked as the initial one
 * (synset originates from the visually similar images).
 * @author Honza Botorek
 */
public class SynsetVertex {
    private boolean isInit;
    private double vertexProbability;
    private Synset synset;

    public SynsetVertex(boolean isInit, double vertexProbability, Synset synset){
        this.isInit = isInit;
        this.vertexProbability = vertexProbability;
        this.synset = synset;
    }

    public boolean isIsInit() {
        return isInit;
    }

    public void setIsInit(boolean isInit) {
        this.isInit = isInit;
    }

    public double getVertexProbability() {
        return vertexProbability;
    }

    public void setVertexProbability(double vertexProbability) {
        this.vertexProbability = vertexProbability;
    }

    public Synset getSynset() {
        return synset;
    }

    public void setSynset(Synset synset) {
        this.synset = synset;
    }

    public String getSynsetWord() {
        Word word = synset.getWord(0);
        if (word == null)
            return synset.toString();
        return word.getLemma();
    }

    @Override
    public int hashCode() {
        int hash = 5;
        return 71 * hash + Objects.hashCode(this.synset);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        SynsetVertex other = (SynsetVertex) obj;
        return this.synset.equals(other.synset);
    }

    @Override
    public String toString() {
        return synset + "   :   " + vertexProbability;
    }

    /**
     * Sort a collection of vertices using the {@link #getVertexProbability() probability}.
     * @param vertices the collection of vertices to sort
     * @return a sorted list of vertices
     */
    public static List<SynsetVertex> sortByProbability(Collection<? extends SynsetVertex> vertices) {
        // Get vertices with the highest probability first
        List<SynsetVertex> ret = new ArrayList<>(vertices);
        Collections.sort(ret, new Comparator<SynsetVertex>() {
            @Override
            public int compare(SynsetVertex o1, SynsetVertex o2) {
                return Double.compare(o2.getVertexProbability(), o1.getVertexProbability()); // Sorting is descending
            }
        });
        return ret;
    }
}
