package messif.objects.annotation.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import messif.objects.annotation.Annotation;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.api.AnnotationTransformer;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;

/**
 * Reducer that merges the {@link AnnotationTerm}s with the same word into one
 * {@link AnnotationTerm} and computes the aggregate confidence.
 *
 * @author Michal Batko <batko@fi.muni.cz>
 */
public abstract class AbstractTermMergeReducer implements AnnotationTransformer<Annotation, Annotation> {
    /**
     * Internal helper class for storing the merged terms.
     */
    private static class ReducedTermData {
        private final String word;
        private boolean userSpecified;
        private final List<Float> confidences = new ArrayList<>();
        private final List<Object> origins = new ArrayList<>();
        private ReducedTermData(String word) {
            this.word = word;
            this.userSpecified = false;
        }
        public void add(float confidence, AnnotationTerm term) {
            this.confidences.add(confidence);
            this.origins.addAll(term.getOriginIdentifiers());
            if (term.isUserSpecified())
                userSpecified = true;
        }
    }

    @Override
    public Annotation classify(Annotation inputAnnotation, Parametric parameters) throws ClassificationException {
        Map<String, ReducedTermData> termDataMap = new HashMap<>();
        for (AnnotationTerm inputTerm : inputAnnotation) {
            ReducedTermData termData = termDataMap.get(inputTerm.getWord());
            if (termData == null) {
                termData = new ReducedTermData(inputTerm.getWord());
                termDataMap.put(inputTerm.getWord(), termData);
            }
            termData.add(inputAnnotation.getConfidence(inputTerm), inputTerm);
        }

        Annotation result = new Annotation(inputAnnotation.getAnnotatedObject());
        for (ReducedTermData termData : termDataMap.values()) {
            result.updateConfidence(new AnnotationTerm(termData.word, termData.userSpecified, termData.confidences.size(), termData.origins), computeAggregateConfidence(termData.confidences));
        }
        return result;
    }

    /**
     * Compute aggregated confidence for the given list of confidences.
     * @param confidences the list of confidences from which to compute the aggregate
     * @return the aggregated confidence
     */
    protected abstract float computeAggregateConfidence(List<Float> confidences);

    @Override
    public Class<? extends Annotation> getAnnotationInputClass() {
        return Annotation.class;
    }

    @Override
    public Class<? extends Annotation> getAnnotationOutputClass() {
        return Annotation.class;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }
}
