package messif.objects.annotation.impl;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import messif.objects.annotation.Annotation;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationWithSynsets;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetVertex;
import messif.objects.annotation.api.AnnotationTransformer;
import messif.objects.annotation.util.WordNetFrequencyManager;
import messif.objects.annotation.util.WordNetOperations;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;
import net.didion.jwnl.data.Synset;

/**
 * Transformer component that transforms input Annotation terms into baseline
 * semantic network; the transformer produce vertex objects. 
 * 
 * Words are transformed into synsets; number of synsets extracted from the particular word
 * depends on defined constant. 
 * 
 * Each synset is furthermore transformed into the vertex with an initial probability.
 * @author Honza Botorek
 */
public class AnnotationTermToNetworkBaseTransformer implements AnnotationTransformer<Annotation, AnnotationWithSynsetsEdges> {
    private final WordNetFrequencyManager wordNetSynsetFreqManager;
    private final int constWnMostProbabSyns;
    private final double wnProbabToConfidenceRatio;
    private final WordNetOperations wnAgent; 

    public AnnotationTermToNetworkBaseTransformer(int constWnMostProbabSyns, double wnProbabToConfidenceRatio, WordNetOperations wnAgent) throws IOException {
        this.wordNetSynsetFreqManager = new WordNetFrequencyManager(new File(wnAgent.getDictionaryPath(), "index.sense"));
        this.constWnMostProbabSyns = constWnMostProbabSyns;
        this.wnProbabToConfidenceRatio = wnProbabToConfidenceRatio;
        this.wnAgent = wnAgent;
    }

    @Override
    public AnnotationWithSynsetsEdges classify(Annotation inputAnnotation, Parametric parameters) throws ClassificationException {
        AnnotationWithSynsets workingAnnotation = new AnnotationWithSynsets(inputAnnotation);
        List<SynsetVertex> vertices = new ArrayList<>();
        
        for(AnnotationTerm at : workingAnnotation){
            String cwText = at.getWord();
            Map<Synset, Double> origWeight = wordNetSynsetFreqManager.selectMostFrequentNormalized(
                    wnAgent.getAllSynsets(cwText),
                    constWnMostProbabSyns
            );
            Collection<Synset> selectedSynsets = origWeight.keySet();
            //List<Synset> selectedSynsets = wordNetSynsetFreqManager.getMostProbableSynsets(wnAgent.getAllSynsets(cwText), constWnMostProbabSyns);
            workingAnnotation.addTermSynsets(at, selectedSynsets);
            for(Synset syn: selectedSynsets){       
                //double wnProbablity = wordNetSynsetFreqManager.getSynsetProbability(syn) * wnProbabToConfidenceRatio;
                double wnProbablity = origWeight.get(syn);
                double synsetProbability;
                if (wnProbabToConfidenceRatio <= 0.01)
                    synsetProbability = inputAnnotation.getConfidence(at);
                else if (wnProbabToConfidenceRatio >= 1.99) {
                    synsetProbability = wnProbablity;
                } else {
                    synsetProbability = wnProbablity * inputAnnotation.getConfidence(at);
                }
                SynsetVertex synVertex = new SynsetVertex(true, synsetProbability , syn);
                int indexOfSynVertex = vertices.indexOf(synVertex);
                if (indexOfSynVertex != -1) {
                    double currentProbabValue = vertices.get(indexOfSynVertex).getVertexProbability();
                    double toAddProbab = synVertex.getVertexProbability();
                    double added = currentProbabValue + toAddProbab; 
                    synVertex.setVertexProbability(added);
                    vertices.remove(indexOfSynVertex);
                }
                vertices.add(synVertex);
            }
        }

        AnnotationWithSynsetsEdges resultAnnotation = new AnnotationWithSynsetsEdges(
                inputAnnotation,
                workingAnnotation.getTermSynsets()
        );
        
        double checkSum = 0;
        for(SynsetVertex sv : vertices){
            checkSum += sv.getVertexProbability();
        }
        if (checkSum != 1) {
            //synsets were not found for all keywords from extendedClassification, therefore the sum of probabilities is not 1 and we need to re-normalize
            for(SynsetVertex sv : vertices) {
                sv.setVertexProbability(sv.getVertexProbability() / checkSum);
            }
        }
        for(SynsetVertex sv : vertices){
            resultAnnotation.addVertexToNetwork(sv);
        }
        
        return resultAnnotation;        
    }

    @Override
    public Class<? extends Annotation> getAnnotationInputClass() {
        return Annotation.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationOutputClass() {
        return AnnotationWithSynsetsEdges.class;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }
    
}
