package messif.objects.annotation.impl;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.regex.Pattern;
import messif.objects.LocalAbstractObject;
import messif.objects.annotation.Annotation;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationTerm.AnnotationTermWithOrigin;
import messif.objects.annotation.api.Annotator;
import messif.objects.classification.ClassificationException;
import messif.objects.classification.Classifier;
import messif.objects.util.RankedAbstractObject;
import messif.operations.RankingMultiQueryOperation;
import messif.operations.RankingQueryOperation;
import messif.operations.RankingSingleQueryOperation;
import messif.utility.ExtendedDatabaseConnection;
import messif.utility.Parametric;

/**
 * Provides a {@link Classifier} that processes list of ranked abstract objects and provides keyword classification
 * with words belonging to the object locator URI with the respective confidence based on the ranking distance.
 * @author Michal Batko <batko@fi.muni.cz>
 */
public class DatabaseKeywordAnnotationClassifier extends ExtendedDatabaseConnection implements Annotator<RankingQueryOperation, Annotation> {
    /** Class id for serialization. */
    private static final long serialVersionUID = 1L;

    /** SQL command to retrieve word links */
    private final String objectLocatorToWordSQL;
    /** Prepared SQL command to retrieve word links */
    private PreparedStatement getObjectWordLinks;
    /** Splitting regular expression for words returned by the database */
    private final Pattern splitWordRegexp;

    /**
     * Creates a database word conversion classifier.
     *
     * @param dbConnUrl the database connection URL (e.g. "jdbc:mysql://localhost/somedb")
     * @param dbConnInfo additional parameters of the connection (e.g. "user" and "password")
     * @param dbDriverClass class of the database driver to use (can be <tt>null</tt> if the driver is already registered)
     * @param objectLocatorToWordSQL the SQL command used to transform the object locator URI (SQL command parameter) to associated words (the returned result-set first column)
     * @param splitRegexp regular expression for splitting words returned by the database
     * @throws IllegalArgumentException if the connection url is <tt>null</tt> or the driver class cannot be registered
     * @throws SQLException if there was a problem connecting to the database
     */
    public DatabaseKeywordAnnotationClassifier(String dbConnUrl, Properties dbConnInfo, String dbDriverClass, String objectLocatorToWordSQL, String splitRegexp) throws IllegalArgumentException, SQLException {
        super(dbConnUrl, dbConnInfo, dbDriverClass);
        this.objectLocatorToWordSQL = objectLocatorToWordSQL;
        getObjectWordLinks = getConnection().prepareStatement(objectLocatorToWordSQL);
        this.splitWordRegexp = splitRegexp == null ? null : Pattern.compile(splitRegexp);
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }

    @Override
    public Class<? extends RankingQueryOperation> getAnnotationInputClass() {
        return RankingQueryOperation.class;
    }

    @Override
    public Class<? extends Annotation> getAnnotationOutputClass() {
        return Annotation.class;
    }

    /**
     * Returns the query object from the operation.
     * @param operation the operation the query object of which to get
     * @return the query object from the operation
     */
    protected LocalAbstractObject getAnnotatedObject(RankingQueryOperation operation) {
        if (operation instanceof RankingSingleQueryOperation)
            return ((RankingSingleQueryOperation)operation).getQueryObject();
        else if (operation instanceof RankingMultiQueryOperation)
            return ((RankingMultiQueryOperation)operation).getQueryObject(0);
        else
            return null;
    }

    /**
     * Converts the distance from the operation answer to the confidence score for the annotation.
     * @param operation the operation for which to transform the results
     * @param result the result object to get the distance for
     * @return the score of the given result
     */
    protected float distanceToScore(RankingQueryOperation operation, RankedAbstractObject result) {
        float maxAnswerDistance = operation.getAnswerDistance();
        return (maxAnswerDistance - result.getDistance()) / maxAnswerDistance;
    }

    /**
     * Returns the words for a given object identifier.
     * This method contacts the database and retrieves the list of rows that provide
     * the words. They can be further split using the {@link #splitWordRegexp}.
     * @param objectLocator the object identifier for which to retrieve words
     * @return a list of words
     * @throws SQLException if there was an error retrieving the data from the database
     */
    protected List<String> getWordsForLocator(String objectLocator) throws SQLException {
        List<String> words = new ArrayList<>();
        synchronized (objectLocatorToWordSQL) {
            getObjectWordLinks = prepareAndExecute(getObjectWordLinks, objectLocatorToWordSQL, false, objectLocator);
            ResultSet rs = getObjectWordLinks.getResultSet();
            while (rs.next()) {
                if (splitWordRegexp == null) {
                    words.add(rs.getString(1));
                } else {
                    words.addAll(Arrays.asList(splitWordRegexp.split(rs.getString(1))));
                }
            }
            rs.close();
        }
        return words;
    }

    /**
     * Update the annotation with the results of the given object.
     * @param annotation the annotation the update
     * @param operation the operation for which to update the annotation
     * @param result the result for which to update the annotation
     * @throws SQLException if there was an error retrieving the data from the database
     */
    protected void updateAnnotationWithResult(Annotation annotation, RankingQueryOperation operation, RankedAbstractObject result) throws SQLException {
        String objectLocator = result.getObject().getLocatorURI();
        float confidence = distanceToScore(operation, result);
        for (String word : getWordsForLocator(objectLocator)) {
            annotation.updateConfidence(new AnnotationTermWithOrigin(word, false, objectLocator), confidence);
        }
    }

    @Override
    public Annotation classify(RankingQueryOperation operation, Parametric parameters) throws ClassificationException {
        Annotation ret = new Annotation(getAnnotatedObject(operation));
        try {
            Iterator<RankedAbstractObject> answerIterator = operation.getAnswer();
            while (answerIterator.hasNext()) {
                updateAnnotationWithResult(ret, operation, answerIterator.next());
            }
        } catch (SQLException e) {
            throw new ClassificationException("Error reading keywords from the database", e);
        }

        return ret;
    }
}
