package messif.objects.annotation.impl;

import java.sql.SQLException;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import messif.objects.annotation.Annotation;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.classification.Classifier;
import messif.objects.util.RankedAbstractObject;
import messif.operations.RankingQueryOperation;

/**
 * Provides a {@link Classifier} that processes list of ranked abstract objects and provides keyword classification
 * with words belonging to the object locator URI with the respective confidence based on the ranking distance.
 * @author Michal Batko <batko@fi.muni.cz>
 */
public class DatabaseKeywordWithRerankingAnnotationClassifier extends DatabaseKeywordAnnotationClassifier {
    /** Class id for serialization. */
    private static final long serialVersionUID = 1L;

    public DatabaseKeywordWithRerankingAnnotationClassifier(String dbConnUrl, Properties dbConnInfo, String dbDriverClass, String objectLocatorToWordSQL, String splitRegexp) throws IllegalArgumentException, SQLException {
        super(dbConnUrl, dbConnInfo, dbDriverClass, objectLocatorToWordSQL, splitRegexp);
    }

    protected Map<String, Float> parseQueryObjectWords(String queryObjectWords) {
        Map<String, Float> queryWeights = new HashMap<>();
        for (String wordWithWeight : queryObjectWords.split(";")) {
            int colonPos = wordWithWeight.lastIndexOf(':');
            if (colonPos == -1) {
                queryWeights.put(wordWithWeight, 1f);
            } else {
                queryWeights.put(wordWithWeight.substring(0, colonPos), Float.parseFloat(wordWithWeight.substring(colonPos + 1)));
            }
        }
        return queryWeights;
    }

    protected float computeTfIdfConfidence(Map<String, Float> queryWeights, List<String> databaseObjectWords) {
        if (databaseObjectWords.isEmpty())
            return 0;

        // Sum of multiplied weights that are in both vectors (i.e. dot product of the two sparse vectors)
        float nominator = 0;
        for (String databaseObjectWord : databaseObjectWords) {
            Float weight = queryWeights.get(databaseObjectWord);
            if (weight != null)
                nominator += weight;
        }

        // Compute vector size for the query words
        double queryWeightsSquareSum = 0;
        for (Float weight : queryWeights.values()) {
            queryWeightsSquareSum += weight * weight;
        }

        // Compute vector size for the database words (all words present have 1 weight)
        double databaseWeightsSquareSum = databaseObjectWords.size();
        // Return scalar vercotr
        return (float)(nominator / (Math.sqrt(queryWeightsSquareSum) * Math.sqrt(databaseWeightsSquareSum)));
    }

    @Override
    protected void updateAnnotationWithResult(Annotation annotation, RankingQueryOperation operation, RankedAbstractObject result) throws SQLException {
        String objectLocator = result.getObject().getLocatorURI();
        List<String> words = getWordsForLocator(objectLocator);
        String queryObjectWords = operation.getParameter("keywords", String.class);
        float confidence = queryObjectWords == null || queryObjectWords.isEmpty() ?
                distanceToScore(operation, result) : computeTfIdfConfidence(parseQueryObjectWords(queryObjectWords), words);
        for (String word : words) {
            annotation.updateConfidence(new AnnotationTerm.AnnotationTermWithOrigin(word, false, objectLocator), confidence);
        }
    }

}
