package messif.objects.annotation.impl;

import Jama.Matrix;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetEdge;
import messif.objects.annotation.SynsetRelationType;
import messif.objects.annotation.SynsetVertex;
import messif.objects.annotation.api.AnnotationTransformer;
import messif.objects.annotation.util.MatrixOperations;
import messif.objects.annotation.util.OwlWorker;
import messif.objects.annotation.util.RelationsValueComputer;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;

/**
 * Transformer component that performs matrix operations representing modified
 * approach to the page ranking algorithm. Operations are performed on the
 * semantic network data.
 * 
 * Component utilizes probability values of particular vertices that are
 * propagated through the network by edges of a certain weights.
 * @author Petra Budíková, Honza Botorek
 */
public class MatrixComputationTransformer implements AnnotationTransformer<AnnotationWithSynsetsEdges, AnnotationWithSynsetsEdges> {
    private final MatrixOperations matrixOp;
    private final List<SynsetRelationType> extenderRelations;
    private final List<SynsetRelationType> relationsToEnhance;
    private final double RESTART_RATIO;
    private final RelationsValueComputer relationsValueComputer;
    
    public MatrixComputationTransformer(OwlWorker owlWorker,
            double restartRatio,
            double one,
            double two,
            SynsetRelationType[] extenderRelations,
            SynsetRelationType[] enhRelations,
            RelationsValueComputer relValComp){
        matrixOp = new MatrixOperations(owlWorker, one, two);
        this.extenderRelations = Arrays.asList(extenderRelations);
        this.relationsToEnhance = Arrays.asList(enhRelations);
        RESTART_RATIO = restartRatio;
        relationsValueComputer = relValComp;
    }

    @Override
    public AnnotationWithSynsetsEdges classify(AnnotationWithSynsetsEdges sourceAnnotation, Parametric parameters) throws ClassificationException {
        Map<SynsetVertex,List<SynsetEdge>> semanticNetwork = sourceAnnotation.getSemanticNetwork();
        Matrix transferM = constructTransferMatrix(semanticNetwork);    
        Matrix absorptionM = matrixOp.constructAbsorptionMatrix(sourceAnnotation);
        Matrix restartM = matrixOp.constructRestartMatrix(sourceAnnotation);
            
        Matrix annotNetwork = matrixOp.createAnnotationNetwork(RESTART_RATIO, restartM, absorptionM, transferM);
               
        int eigenVectorPointer = getPrimaryEigenValue(annotNetwork);            // 
        if ( eigenVectorPointer != -1) {
            Matrix annotationNetworkAllVector1 = annotNetwork.eig().getV().getMatrix(0, annotNetwork.eig().getV().getRowDimension()-1, eigenVectorPointer, eigenVectorPointer);
            annotationNetworkAllVector1 = annotationNetworkAllVector1.times(1.0/annotationNetworkAllVector1.norm1());
            if (getColumnSum(annotationNetworkAllVector1)[0] < 0) {
                annotationNetworkAllVector1 = annotationNetworkAllVector1.uminus();
            }            
                        
            double[] eigenValues = annotationNetworkAllVector1.getColumnPackedCopy();
            List<SynsetVertex> vertices = new ArrayList<>(sourceAnnotation.getSemanticNetwork().keySet());
            for(int j = 0; j < eigenValues.length; j++){
                SynsetVertex vertex = vertices.get(j);
                vertex.setVertexProbability(eigenValues[j] * 100);
            }            
            
            return sourceAnnotation;
        } else{
            throw new ClassificationException("EIGEN VECTOR POINTER == -1");
        }       
    }

    /**
     * Method returns primary eigen value - the first eigen vector with the value
     * higher than 92 and lower and equal than 100
     * @param annotNetwork Semantic network matrix representation 
     * @return primary eigen value 
     */
     private int getPrimaryEigenValue(Matrix annotNetwork) {
         for(int i = 0; i < annotNetwork.getColumnDimension(); i++){
             double eigenValue = Math.round(annotNetwork.eig().getRealEigenvalues()[i]*100);             // 
             if ( eigenValue <= 100 && eigenValue > 92) {
                 return i;
             }
         }
        throw new IllegalStateException("Problem - the first eigenvalue is not in the specified bounds ( 92 < X < 100)!");
    }

     /**
      * Method construct transfer matrix from the given semantic network.
      * Transfer matrix contains elements representing edge weights of semantic 
      * relation-based edges connecting particular vertices.
      * @param processedVertices semantic network
      * @return transfer matrix 
      */
    private Matrix constructTransferMatrix(Map<SynsetVertex,List<SynsetEdge>> processedVertices) {
        int dim = processedVertices.size();
        Matrix wholeTransfer = new Matrix(new double[dim][dim]);
        for(SynsetRelationType re : extenderRelations){
            Matrix transposed = new Matrix(new double[dim][dim]);;
            Matrix transfer = matrixOp.constructTypeMatrix(processedVertices, re);
            if(re.equals(SynsetRelationType.WN_HYPERNYM) && relationsToEnhance.contains(SynsetRelationType.WN_HYPONYM)){
                transposed = transfer.transpose();
                transposed = transposed.times(relationsValueComputer.getHyponymConstant());
                transfer = transfer.plus(transposed);
            } else if(re.equals(SynsetRelationType.WN_HOLONYM) && relationsToEnhance.contains(SynsetRelationType.WN_MERONYM)){
                transposed = transfer.transpose();
                transposed = transposed.times(relationsValueComputer.getMeronymConstant());
                transfer = transfer.plus(transposed);
            }
            wholeTransfer = wholeTransfer.plus(transfer);
        }
        return wholeTransfer;
    }
    
    /**
     * Method returns sum of all elements in a single column
     * @param A matrix containing certain columns
     * @return array of sums of elements in particular columns
     */
    private double[] getColumnSum (Matrix A) {
      double[] result = new double[A.getColumnDimension()];      
      for (int j = 0; j < A.getColumnDimension(); j++) {
         double s = 0;
         for (int i = 0; i < A.getRowDimension(); i++) {
            s += A.get(i, j);
         }
         result[j] = s;
      }
      return result;
   }
    
    

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationInputClass() {
        return AnnotationWithSynsetsEdges.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationOutputClass() {
        return AnnotationWithSynsetsEdges.class;
    }
    
}
