
package messif.objects.annotation.impl;

import java.util.ArrayList;
import java.util.List;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetEdge;
import messif.objects.annotation.SynsetRelationType;
import messif.objects.annotation.SynsetVertex;
import messif.objects.annotation.api.AnnotationExpander;
import messif.objects.annotation.util.RelationsValueComputer;
import messif.objects.annotation.util.WordNetOperations;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;
import net.didion.jwnl.data.Synset;

/**
 * Expander component utilizing WordNet relationship - HOLONYMY.
 * 
 * Semantic network is expanded by new vertices (synsets) based on the defined
 * WordNet relationship. Furthermore, when relationship is formed, an edge
 * is added into the network. The edge weight is defined by the relationship type.
 * @author Honza Botorek
 */
public class WNHolonymExtenderComponent implements AnnotationExpander<AnnotationWithSynsetsEdges, AnnotationWithSynsetsEdges> {
    private WordNetOperations wnAgent; 
    private RelationsValueComputer edgeWeightComputer;

    public WNHolonymExtenderComponent(WordNetOperations wnAgent, RelationsValueComputer edgeWeightComputer) {
        this.wnAgent = wnAgent;
        this.edgeWeightComputer = edgeWeightComputer;
    }
    
    
    @Override
    public AnnotationWithSynsetsEdges classify(AnnotationWithSynsetsEdges sourceAnnotation, Parametric prmtrc) throws ClassificationException {
        AnnotationWithSynsetsEdges resultAnnotation = new AnnotationWithSynsetsEdges(sourceAnnotation, sourceAnnotation.getTermSynsets());
        List<SynsetVertex> queue = new ArrayList<>(sourceAnnotation.getSemanticNetwork().keySet());
        while(!queue.isEmpty()){
            SynsetVertex activeVertex = queue.get(0);
            resultAnnotation.copyVertex(activeVertex, sourceAnnotation, resultAnnotation);
            String offsetStringValue = String.valueOf(activeVertex.getSynset().getOffset());
            if(!offsetStringValue.startsWith("99999")){                                       //Virtual synset!!!!
                List<Synset> retrieved = wnAgent.getDirectHolonyms(activeVertex.getSynset());
                for(Synset vertex : retrieved){
                    SynsetVertex retrievedVertex = new SynsetVertex(false, 0, vertex);
                    double edgeWeight = edgeWeightComputer.computeRelationshipValue(SynsetRelationType.WN_HOLONYM, retrieved.size());
                    SynsetEdge edge = new SynsetEdge(edgeWeight, vertex, SynsetRelationType.WN_HOLONYM);
                    resultAnnotation.addSynsetEdgeToVertex(activeVertex, edge);
                    if(!resultAnnotation.networkContainsVertex(activeVertex) && 
                            !queue.contains(retrievedVertex)){
                        queue.add(retrievedVertex);
                    }
                }
            }
            queue.remove(0);          
        }       
        return resultAnnotation;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationInputClass() {
        return AnnotationWithSynsetsEdges.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationOutputClass() {
        return AnnotationWithSynsetsEdges.class;
    }
}
