package messif.objects.annotation.impl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetEdge;
import messif.objects.annotation.SynsetRelationType;
import messif.objects.annotation.SynsetVertex;
import messif.objects.annotation.api.AnnotationExpander;
import messif.objects.annotation.util.OwlWorker;
import messif.objects.annotation.util.RelationsValueComputer;
import messif.objects.annotation.util.WordNetOperations;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.data.Synset;
import net.didion.jwnl.data.Word;

/**
 * Expander component utilizing WordNet relationship - HYPERNYMY.
 * 
 * Semantic network is expanded by new vertices (synsets) based on the defined
 * WordNet relationship. Furthermore, when relationship is formed, an edge
 * is added into the network. The edge weight is defined by the relationship type.
 * @author Honza Botorek
 */
public class WNHypernymExtenderComponent implements AnnotationExpander<AnnotationWithSynsetsEdges, AnnotationWithSynsetsEdges> {
    private WordNetOperations wnAgent; 
    private OwlWorker owlWorker;
    private Set<Long> borderSynsets;
    private RelationsValueComputer edgeWeightComputer;
    
    private int virtualSynsetCounter;
    private long virtualSynsetBase;
    private Map<String,Long> virtualWordToLongMapping;

    public WNHypernymExtenderComponent(WordNetOperations wnAgent,OwlWorker owlWorker, RelationsValueComputer edgeValueComputer) {
        this.wnAgent = wnAgent;
        this.owlWorker = owlWorker;
        initBorderSynsets();
        virtualSynsetCounter = 0;
        virtualSynsetBase = 999990000;
        this.edgeWeightComputer = edgeValueComputer;
        virtualWordToLongMapping = new HashMap<>();
    }
    
    
    @Override
    public AnnotationWithSynsetsEdges classify(AnnotationWithSynsetsEdges sourceAnnotation, Parametric parameters) throws ClassificationException {        
        AnnotationWithSynsetsEdges resultAnnotation = new AnnotationWithSynsetsEdges(sourceAnnotation, sourceAnnotation.getTermSynsets());
        List<SynsetVertex> queue = new ArrayList<>(sourceAnnotation.getSemanticNetwork().keySet());
        while(!queue.isEmpty()){
            SynsetVertex activeVertex = queue.get(0);
            resultAnnotation.copyVertex(activeVertex, sourceAnnotation, resultAnnotation);
            String offsetStringValue = String.valueOf(activeVertex.getSynset().getOffset());
            if(!offsetStringValue.startsWith("99999")){                                       //Virtual synset!!!!
                List<Synset> retrieved = extendWords(activeVertex.getSynset());
                if(retrieved != null){
                    for(Synset synset : retrieved){
                        SynsetVertex retrievedVertex = new SynsetVertex(false, 0, synset);
                        double edgeWeight = edgeWeightComputer.computeRelationshipValue(SynsetRelationType.WN_HYPERNYM, retrieved.size());
                        SynsetEdge edge = new SynsetEdge(edgeWeight, synset, SynsetRelationType.WN_HYPERNYM);
                        resultAnnotation.addSynsetEdgeToVertex(activeVertex, edge);
                        if(!resultAnnotation.networkContainsVertex(retrievedVertex) && 
                                !queue.contains(retrievedVertex)){
                            queue.add(retrievedVertex);                            
                        }
                    }
                }
            }
            queue.remove(0);          
        }   
        return resultAnnotation;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }
    
    private List<Synset> extendWords(Synset activeSynset) {
        List<Synset> hypernyms = new ArrayList<>();
        if(!borderSynsets.contains(activeSynset.getOffset())){
            if(activeSynset.getPOS().equals(POS.VERB)){
                Synset hypernym = wnAgent.getDirectVerbHypernym(activeSynset);
                if(hypernym!= null) hypernyms.add(hypernym);
            } else{
                Synset hypernym = wnAgent.getDirectHypernym(activeSynset);
                if(hypernym != null){
                    if(!borderSynsets.contains(hypernym.getOffset())){
                        hypernyms.add(hypernym);
                    } else{                                 // one of the border synsets
                        hypernyms.add(hypernym);
                    }
                }
            }
        }else {
            List<String> ontologySuperClasses = owlWorker.getSuperCategories(activeSynset);
            if(ontologySuperClasses != null){
                List<Synset> superSynsets = constructSyns(ontologySuperClasses);
                hypernyms.addAll(superSynsets);
            } else hypernyms = null;
        }        
        return hypernyms;
    }

    private void initBorderSynsets() {
        borderSynsets = owlWorker.getLocatorsOfVcoBorderEntities();
        
        borderSynsets.add(wnAgent.getAbstractEntity().getOffset());                     //abstract entity
        borderSynsets.add(wnAgent.getPhysicalEntity().getOffset());                     //physical entity
        borderSynsets.add(wnAgent.getCentralBorderEntity().getOffset());                //central border entity
    }

    private List<Synset> constructSyns(List<String> ontologySuperClasses) {
        List<Synset> virtualSynsets = new ArrayList<>();
        for(String iri : ontologySuperClasses){
            if(!iri.contains("www.w3.org")){                                        // www.w3.org is a namespace of default Protege root of the ontology
                String text = iri.split("/")[4];
                Word word = new Word("VIRT_" + text,"VIRT_" +  text, 1);
                Word[] words = {word};
                long offset = assignVirtualNumber(text);
                Synset syn = new Synset(POS.NOUN, offset, words, null, iri, null);
                virtualSynsets.add(syn);
            }
        }
        return virtualSynsets;
    }

    private long assignVirtualNumber(String text) {
        if(virtualWordToLongMapping.containsKey(text)){
            return virtualWordToLongMapping.get(text);
        } else{
            virtualSynsetCounter++;
            long resultOffset = virtualSynsetCounter + virtualSynsetBase;
            virtualWordToLongMapping.put(text, resultOffset);
            return resultOffset;
            
        }
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationInputClass() {
        return AnnotationWithSynsetsEdges.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationOutputClass() {
        return AnnotationWithSynsetsEdges.class;
    }
}
