package messif.objects.annotation.impl;

import java.util.Arrays;
import java.util.List;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetEdge;
import messif.objects.annotation.SynsetRelationType;
import messif.objects.annotation.SynsetVertex;
import messif.objects.annotation.api.AnnotationTransformer;
import messif.objects.annotation.util.RelationsValueComputer;
import messif.objects.annotation.util.WordNetOperations;
import messif.utility.Parametric;
import net.didion.jwnl.data.Synset;

/**
 * Transformer component utilizing WordNet relationship - HYPONYMY.
 * 
 * Semantic network is enhanced by new edges between particular vertices
 * according to the given WordNet relationship. The edges are formed ONLY 
 * between existing vertices; NO new vertices are addded into the network.
 * 
 * @author Honza Botorek
 */
public class WNHyponymEnhancer implements AnnotationTransformer<AnnotationWithSynsetsEdges, AnnotationWithSynsetsEdges> {
    private final WordNetOperations wnAgent;
    private final RelationsValueComputer edgeWeightComputer;

    public WNHyponymEnhancer(WordNetOperations wnAgent, RelationsValueComputer edgeWeightComputer) {
        this.wnAgent = wnAgent;
        this.edgeWeightComputer = edgeWeightComputer;
    }
    
    @Override
    public AnnotationWithSynsetsEdges classify(AnnotationWithSynsetsEdges sourceAnnotation, Parametric prmtrc){
        //AnnotationWithSynsetsEdges outputAnnotation = new AnnotationWithSynsetsEdges(sourceAnnotation);
        for(SynsetVertex activeVertex : sourceAnnotation.getSemanticNetwork().keySet()){
            String activeSynsetString = String.valueOf(activeVertex.getSynset().getOffset());
            if(!activeSynsetString.startsWith("99999")){
                List<Synset> hyponyms = Arrays.asList(wnAgent.getHyponyms(activeVertex.getSynset()));
                int hyponymsNumber = hyponyms.size();
                for(Synset hyponym : hyponyms){
                    SynsetVertex retrievedVertex = new SynsetVertex(false, 0, hyponym);
                    if(sourceAnnotation.getSemanticNetwork().containsKey(retrievedVertex)){                        
                        double relationValue = edgeWeightComputer.computeRelationshipValue(SynsetRelationType.WN_HYPONYM, hyponymsNumber);
                        SynsetEdge newEdge = new SynsetEdge(relationValue, hyponym, SynsetRelationType.WN_HYPONYM);
                        sourceAnnotation.addSynsetEdgeToVertex(activeVertex, newEdge);
                    }
                }
            }
        }
        return sourceAnnotation;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationInputClass() {
        return AnnotationWithSynsetsEdges.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationOutputClass() {
        return AnnotationWithSynsetsEdges.class;
    }
}
