
package messif.objects.annotation.impl;

import java.util.Arrays;
import java.util.List;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetEdge;
import messif.objects.annotation.SynsetRelationType;
import messif.objects.annotation.SynsetVertex;
import messif.objects.annotation.api.AnnotationTransformer;
import messif.objects.annotation.util.RelationsValueComputer;
import messif.objects.annotation.util.WordNetOperations;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;
import net.didion.jwnl.data.Synset;

/**
 * Transformer component utilizing WordNet relationship - MERONYMY.
 * 
 * Semantic network is enhanced by new edges between particular vertices
 * according to the given WordNet relationship. The edges are formed ONLY 
 * between existing vertices; NO new vertices are addded into the network.
 * 
 * @author Honza Botorek
 */
public class WNMeronymEnhancer implements AnnotationTransformer<AnnotationWithSynsetsEdges, AnnotationWithSynsetsEdges> {
    private final WordNetOperations wnAgent;
    private final RelationsValueComputer edgeWeightComputer;
    
    public WNMeronymEnhancer(WordNetOperations wnAgent, RelationsValueComputer edgeWeightComputer){
        this.wnAgent = wnAgent;
        this.edgeWeightComputer = edgeWeightComputer;
    }
    
    /**
     * Method examines every vertex of the graph - it produces it's meronyms.
     * For each meronym it searches the graph, if it is also presented. If so,
     * relationship (edge) is constructed.
     * Meronyms are not examined for "virtual synsets" - vertices constructed from 
     * the VCO ontology.
     * 
     * 
     * @param sourceAnnotation
     * @param prmtrc
     * @return 
     */
    @Override
    public AnnotationWithSynsetsEdges classify(AnnotationWithSynsetsEdges sourceAnnotation, Parametric prmtrc) throws ClassificationException {
        for(SynsetVertex activeVertex : sourceAnnotation.getSemanticNetwork().keySet()){
            String activeSynsetString = String.valueOf(activeVertex.getSynset().getOffset());
            if(!activeSynsetString.startsWith("99999")){
                List<Synset> meronyms = Arrays.asList(wnAgent.getMeronyms(activeVertex.getSynset()));
                int meronymsNumber = meronyms.size();
                for(Synset meronym : meronyms){                    
                    SynsetVertex retrievedVertex = new SynsetVertex(false, 0, meronym);
                    if(sourceAnnotation.getSemanticNetwork().containsKey(retrievedVertex)){
                        double relationValue = edgeWeightComputer.computeRelationshipValue(SynsetRelationType.WN_MERONYM, meronymsNumber);
                        SynsetEdge newEdge = new SynsetEdge(relationValue, meronym, SynsetRelationType.WN_MERONYM);
                        sourceAnnotation.addSynsetEdgeToVertex(activeVertex, newEdge);
                    }
                }
            }
        }
        return sourceAnnotation;
    }

    @Override
    public Class<? extends AnnotationTerm> getCategoriesClass() {
        return AnnotationTerm.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationInputClass() {
        return AnnotationWithSynsetsEdges.class;
    }

    @Override
    public Class<? extends AnnotationWithSynsetsEdges> getAnnotationOutputClass() {
        return AnnotationWithSynsetsEdges.class;
    }
}
