package messif.objects.annotation.pipeline;

import messif.objects.annotation.Annotation;
import messif.objects.annotation.AnnotationPipeline;
import messif.objects.annotation.AnnotationTerm;
import messif.objects.annotation.api.Annotator;
import messif.objects.classification.ClassificationException;
import messif.utility.Parametric;

/**
 * Provides an annotation pipeline that uses an existing {@link Annotation}
 * stored in a {@link Parametric} object as initial annotator. Then, additional
 * processors can be added as needed.
 * <p>
 * This is intended to be used in tandem with multiple {@link messif.algorithms.impl.ClassifierAlgorithm}s
 * where one provides some annotation, stores it in the operation parameter
 * (see {@link messif.algorithms.impl.ClassifierAlgorithm#classificationParameterName})
 * and another pipeline takes over.
 * </p>
 *
 * @author Michal Batko, Masaryk University, Brno, Czech Republic, batko@fi.muni.cz
 */
public class OperationParameterAnnotationPipeline extends AnnotationPipeline<Parametric> {

    /**
     * Creates a pipeline that uses an initial annotation stored in a {@link Parametric} object.
     * @param parameterName the parameter name where the initial annotation is stored
     * @param parameterClass the class of the object stored in the parameter
     */
    public OperationParameterAnnotationPipeline(final String parameterName, final Class<? extends Annotation> parameterClass) {
        super(new Annotator<Parametric, Annotation>() {
            @Override
            public Annotation classify(Parametric object, Parametric parameters) throws ClassificationException {
                return object.getRequiredParameter(parameterName, parameterClass);
            }
            @Override
            public Class<? extends Parametric> getAnnotationInputClass() {
                return Parametric.class;
            }
            @Override
            public Class<? extends Annotation> getAnnotationOutputClass() {
                return parameterClass;
            }
            @Override
            public Class<? extends AnnotationTerm> getCategoriesClass() {
                return AnnotationTerm.class;
            }
        });
    }
}
