package messif.objects.annotation.pipeline;

import java.net.UnknownHostException;
import java.sql.SQLException;
import messif.algorithms.AlgorithmMethodException;
import messif.algorithms.RMIAlgorithm;
import messif.objects.annotation.Annotation;
import messif.objects.annotation.AnnotationPipeline;
import messif.objects.annotation.api.Annotator;
import messif.objects.annotation.impl.DatabaseKeywordAnnotationClassifier;
import messif.objects.annotation.impl.DatabaseKeywordWithRerankingAnnotationClassifier;
import messif.objects.classification.ClassificationException;
import messif.operations.RankingQueryOperation;
import messif.utility.Parametric;

/**
 * Provides an annotation via executing a pipeline of several annotation classifiers.
 * The annotation initializer is a {@link DatabaseKeywordAnnotationClassifier}.
 * The input operation is first executed by the given RMI algorithm then the
 * operation answer is transformed to the initial annotation by database keyword lookup.
 *
 * @author Michal Batko, Masaryk University, Brno, Czech Republic, batko@fi.muni.cz
 */
public class RankingQueryOperationWithDBAnnotationPipeline extends AnnotationPipeline<RankingQueryOperation> {
    /** Algorithm using which the operations are executed */
    private final RMIAlgorithm algorithm;

    /**
     * Creates a new annotation pipeline.
     * @param algorithmHost the IP address or host name of the running algorithm with RMI
     * @param algorithmPort the port of the running algorithm RMI
     * @param dbConnUrl the database connection URL to use for keyword lookup
     * @param objectLocatorToWordSQL the SQL command used to transform the object locator URI (SQL command parameter) to associated words (the returned result-set first column)
     * @param splitRegexp regular expression for splitting words returned by the database
     * @param useReranking flag specifying whether to use the {@link DatabaseKeywordWithRerankingAnnotationClassifier} (<tt>true</tt>) or
     *          the {@link DatabaseKeywordAnnotationClassifier} (<tt>false</tt>)
     * @throws IllegalArgumentException if the connection url is <tt>null</tt> or the driver class cannot be registered
     * @throws SQLException if there was a problem connecting to the database
     * @throws UnknownHostException if the host name cannot be resolved to IP address
     */
    public RankingQueryOperationWithDBAnnotationPipeline(String algorithmHost, int algorithmPort, String dbConnUrl, String objectLocatorToWordSQL, String splitRegexp, boolean useReranking) throws IllegalArgumentException, SQLException, UnknownHostException {
        this(algorithmHost, algorithmPort, useReranking ?
                new DatabaseKeywordWithRerankingAnnotationClassifier(dbConnUrl, null, null, objectLocatorToWordSQL, splitRegexp) :
                new DatabaseKeywordAnnotationClassifier(dbConnUrl, null, null, objectLocatorToWordSQL, splitRegexp)
        );
    }

    /**
     * Creates a new annotation pipeline.
     * @param algorithmHost the IP address or host name of the running algorithm with RMI
     * @param algorithmPort the port of the running algorithm RMI
     * @param initialAnnotator the annotator that provides the initial annotation that will be processed by the pipeline
     * @throws IllegalArgumentException if the connection url is <tt>null</tt> or the driver class cannot be registered
     * @throws SQLException if there was a problem connecting to the database
     * @throws UnknownHostException if the host name cannot be resolved to IP address
     */
    public RankingQueryOperationWithDBAnnotationPipeline(String algorithmHost, int algorithmPort, Annotator<RankingQueryOperation, ?> initialAnnotator) throws IllegalArgumentException, SQLException, UnknownHostException {
        super(initialAnnotator);
        this.algorithm = algorithmHost == null ? null : new RMIAlgorithm(algorithmHost, algorithmPort);
    }

    @Override
    public Annotation classify(RankingQueryOperation operation, Parametric parameters) throws ClassificationException {
        try {
            return super.classify(algorithm == null ? operation : algorithm.executeOperation(operation), parameters);
        } catch (AlgorithmMethodException e) {
            throw new ClassificationException("Error executing operation: " + e, e);
        } catch (NoSuchMethodException e) {
            throw new ClassificationException(e.getMessage(), e);
        }
    }

}
