
package messif.objects.annotation.util;

import Jama.Matrix;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import messif.objects.annotation.AnnotationWithSynsetsEdges;
import messif.objects.annotation.SynsetEdge;
import messif.objects.annotation.SynsetRelationType;
import messif.objects.annotation.SynsetVertex;

/**
 * Helper method enabling matrix operations; it supports the matrix processing
 * transformer component (MatrixComputationTransformer).
 * @author Honza Botorek
 */
public class MatrixOperations {
    private final double ABSORPTION_1ST_LEVEL;
    private final double ABSORPTION_2ST_LEVEL;
    private final OwlWorker owlWorker;

    public MatrixOperations(OwlWorker owlWorker, double absorptionFirstLevel, double absorptionSecondLevel) {
        ABSORPTION_1ST_LEVEL = absorptionFirstLevel;
        ABSORPTION_2ST_LEVEL = absorptionSecondLevel;
        this.owlWorker = owlWorker;
    }
    
    /**
     * Method constructs restart matrix from given graph vertices.
     * Restart matrix = vertices constructed from initial words(synsets) are assigned
     * their initial probability value - other synsets (derived by WN relations) are assigned
     * 0 as a probability value.
     * @param sourceAnnotation object that contains semantic network
     * @return 
     */
    public Matrix constructRestartMatrix(AnnotationWithSynsetsEdges sourceAnnotation){
        double[] vector = new double[sourceAnnotation.getSemanticNetwork().size()];
        List<SynsetVertex> vertices = new ArrayList<>(sourceAnnotation.getSemanticNetwork().keySet());
        for(int i = 0; i < vector.length; i++){
            SynsetVertex vertex = vertices.get(i);
            boolean isInit = vertices.get(i).isIsInit();
            if(isInit){
                vector[i] = vertex.getVertexProbability();
            } else{
                vector[i] = 0;
            }
        }
        Matrix restartVector = new Matrix(new double[][]{ vector });
        Matrix unityVector = new Matrix(1, sourceAnnotation.getSemanticNetwork().size(), 1);
        
        Matrix restartMatrix = restartVector.transpose().times(unityVector);
      //      restartMatrix.print(7, 3);
        return restartMatrix;
    }
    
    /**
     * Method constructs matrix by given relation type.
     * @param vertices semantic network
     * @param relationType type of the defined semantic relationship that is used 
     * to construct the values of matrix elements
     * @return traversal matrix of a certain type
     */
    public Matrix constructTypeMatrix(Map<SynsetVertex,List<SynsetEdge>> vertices, SynsetRelationType relationType){
        double[][] matrixArrays = new double[vertices.size()][vertices.size()];
        List<SynsetVertex> keySynsets = new ArrayList<>(vertices.keySet());
        for(int i=0; i<keySynsets.size(); i++){
            List<SynsetEdge> activeEdges = vertices.get(keySynsets.get(i));
            Map<Long,Double> relatedEdges = getRelatedEdges(activeEdges,relationType);
            for(int j = 0; j < keySynsets.size(); j++){
                long otherVertexOffset = keySynsets.get(j).getSynset().getOffset();
                if(relatedEdges.containsKey(otherVertexOffset)){
                    matrixArrays[i][j] = relatedEdges.get(otherVertexOffset);                           // input probab value into the map
                } else{
                    matrixArrays[i][j] = 0;
                }
            }
        }
        Matrix matrix = new Matrix(matrixArrays);
           // matrix.print(7, 3);
        matrix = matrix.transpose();
        return matrix;
    }

    private Map<Long,Double> getRelatedEdges(List<SynsetEdge> activeEdges, SynsetRelationType relationType) {
        Map<Long,Double> selectedEdgesMap = new HashMap<>();
        for(SynsetEdge edge : activeEdges){
            if(edge.getEdgeType().equals(relationType)){
                selectedEdgesMap.put(edge.getEndVertexLocator(),edge.getEdgeValue());
            }
        }
        return selectedEdgesMap;
    }
    
    /**
     * Method constructs an absorption matrix.
     * Absorption matrix: Matrix that is constructed from a vector
     * containing zero values on all positions except the ones where it is
     * desired to terminate the network traversal (the vertices are too general).
     * In current implementation, absorption vertices are the ones belonging
     * into the VCO.
     * @param sourceAnnotation object that contains semantic network
     * @return absorption matrix
     */
    public Matrix constructAbsorptionMatrix(AnnotationWithSynsetsEdges sourceAnnotation){
        int matrixDim = sourceAnnotation.getSemanticNetwork().size();
        double[] vector = new double[matrixDim];
        List<SynsetVertex> keySynsets = new ArrayList<>(sourceAnnotation.getSemanticNetwork().keySet());                
        for(int i = 0; i < matrixDim; i++){
            String offsetString = String.valueOf(keySynsets.get(i).getSynset().getOffset());
            if(offsetString.startsWith("99999")){
                String vcoExploreString = keySynsets.get(i).getSynset().getWords()[0].getLemma();
                String filteredVcoExploreString = vcoExploreString.replace("VIRT_", "");
                int level = owlWorker.getLevel(filteredVcoExploreString);
                if(level == 1){
                    vector[i] = ABSORPTION_1ST_LEVEL;
                } else if( level == 2){
                    vector[i] = ABSORPTION_2ST_LEVEL;
                } 
                else{
                    vector[i] = 0;
                }
            } else{
                vector[i] = 0;
            }                      
        }
        
        Matrix absorptionVector = new Matrix(new double[][]{ vector } );
        Matrix diagonalAbsorptionMatrix = createDiagonalMatrixFromVector(absorptionVector);
            //diagonalAbsorptionMatrix.print(7, 3);
        return diagonalAbsorptionMatrix;
    }
    
    /**
     * Creates the matrix for the annotation network modelled as Markov system
     *
     * @param restartProbability the probabiliy of random restart
     * @param restartMatrix stochastic matrix of random restart probabilities
     * @param absorptionMatrix stochastic matrix of absorption probabilities
     * @param transferMatrix matrix of transfer probabilities - does not have to be stochastic, will be normalized
     * @return annotation network matrix
     * @throws IllegalArgumentException if <ul><li>the input matrices are not of the same size</li>
     *                                         <li>the restart matrix or absorption matrix are not stochastic</li>
     *                                         <li>the restart probability is out of probability range</li>
     */
   public Matrix createAnnotationNetwork(double restartProbability, Matrix restartMatrix, Matrix absorptionMatrix, Matrix transferMatrix) {
       //check the inputs
       if (!sameSizeMatrix(restartMatrix, absorptionMatrix) || !sameSizeMatrix(restartMatrix, transferMatrix)) {
           throw new IllegalArgumentException("The input matrices are not the same size!");
       }
       if (restartMatrix.norm1()>1.2 || absorptionMatrix.norm1()>1.2) {
           throw new IllegalArgumentException("The restart matrix and absorption matrix need to be stochastic!");
       }
       if (restartProbability<0 || restartProbability>1) {
           throw new IllegalArgumentException("Restart probability has to be a number between 0 and 1!");
       }
       
       //start with the restart matrix, multiplied by restart probability
       Matrix result = restartMatrix.times(restartProbability);
       
       //normalize the transfer matrix so that the sum of each column is 1-diagonal_value of the absorption matrix
       //if an empty column is found in the transfer matrix, set appropriate absorption coefficient to 1
       
       //lets check each column
       for (int i=0; i<transferMatrix.getColumnDimension(); i++) {                      
           if (transferMatrix.getMatrix(0, transferMatrix.getRowDimension()-1, i, i).norm1() != 0) {
               //column non-empty, lets normalize
               double normalizationFactor = (1-absorptionMatrix.get(i, i))/transferMatrix.getMatrix(0, transferMatrix.getRowDimension()-1, i, i).norm1();
               //now we can normalize values in this column
               for (int j=0; j<transferMatrix.getRowDimension(); j++) {
                   transferMatrix.set(j, i, transferMatrix.get(j, i)*normalizationFactor);
               }
           } else {               
               //empty column, all will be absorbed
               absorptionMatrix.set(i, i, 1);
           }           
       }
       //add the normalized transfer matrix and absorption matrix to result (mutliplied by 1-restartProbability)
       result = result.plus(absorptionMatrix.times(1-restartProbability));
       result = result.plus(transferMatrix.times(1-restartProbability));
       return result;
   }
   
   /**
    * Method checks whether or not are two provided matrices of a same dimensions.
    * @param A matrix A
    * @param B matrix B
    * @return true if two matrices are of the same dimensions; false otherwise
    */
   private static boolean sameSizeMatrix(Matrix A, Matrix B) {
       return (A.getColumnDimension() == B.getColumnDimension() && A.getRowDimension() == B.getRowDimension());
   }
   
   /**
    * Method used when the absorption matrix is constructed - from given vector,
    * values in the vector are passed into the main diagonal.
    * @param A
    * @return 
    */
   private Matrix createDiagonalMatrixFromVector(Matrix A) {
      if (A.getColumnDimension() != 1 && A.getRowDimension() != 1) {
          throw new IllegalArgumentException("The input matrix is not a vector!");
      }
      
      if (A.getRowDimension() != 1) {
          A = A.transpose();
      }
      Matrix result = new Matrix(A.getColumnDimension(),A.getColumnDimension());
      for (int i=0; i<A.getColumnDimension(); i++) {
          result.set(i, i, A.get(0, i));
      }
      return result;
   }

}
