package messif.objects.annotation.util;

import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.data.Synset;
import org.semanticweb.owlapi.apibinding.OWLManager;
import org.semanticweb.owlapi.model.IRI;
import org.semanticweb.owlapi.model.OWLClass;
import org.semanticweb.owlapi.model.OWLClassExpression;
import org.semanticweb.owlapi.model.OWLIndividual;
import org.semanticweb.owlapi.model.OWLOntology;
import org.semanticweb.owlapi.model.OWLOntologyCreationException;
import org.semanticweb.owlapi.model.OWLOntologyManager;
import org.semanticweb.owlapi.reasoner.OWLReasoner;
import org.semanticweb.owlapi.reasoner.OWLReasonerFactory;
import org.semanticweb.owlapi.vocab.OWLRDFVocabulary;

/**
 * Helper class allowing manipulation with ontologies. It contains method 
 * processing the VCO ontology in particular.
 * @author Honza Botorek
 */
public class OwlWorker {
    OWLOntologyManager owlMan;
    OWLOntology ontology;
    Map<Long,OWLClass> offsetToClassNameMapping;
    private OWLReasoner reasoner;

    public OwlWorker(String vcoOntologyPath) {
        if(vcoOntologyPath != null){
            owlMan = OWLManager.createOWLOntologyManager();
            offsetToClassNameMapping = new HashMap<>();
            try {
                ontology = owlMan.loadOntologyFromOntologyDocument(new File(vcoOntologyPath));
                OWLReasonerFactory reasonerFactory = (OWLReasonerFactory) Class.forName("com.clarkparsia.pellet.owlapiv3.PelletReasonerFactory").newInstance();
                reasoner = reasonerFactory.createNonBufferingReasoner(ontology); 
            } catch (OWLOntologyCreationException | IllegalAccessException | ClassNotFoundException | InstantiationException ex) {
                throw new IllegalStateException(ex);
            }
        }
    }
    

    /**
     * Method returns a string that expresses synset's presence in the VCO ontology.
     * The string contains all category nodes to be traversed for one to be 
     * able to access the class where the input synset belongs into.
     * Method returns empty string if the ontology is not inicialized.
     * @param synsetById identification of the synset to be sorted into the
     * particular category.
     * @return string that expresses synset's presence in the VCO ontology
     * or returns empty string if the ontology is not inicialized.
     */
    public String checkCategories(Synset synsetById) {
        if(ontology != null){
            long offsetSyn = synsetById.getOffset();
            for(OWLClass cls : ontology.getClassesInSignature()){
                Set<OWLIndividual> individs = cls.getIndividuals(ontology);
                for(OWLIndividual indiv : individs){
                    String offsS = indiv.toString().split("-")[3];
                    long offs = Long.parseLong(offsS);
                    if(offs == offsetSyn){
                        String categoryPath = constructCategoryPath(cls);
                        return categoryPath;
                    }
                }

            }
        } else{
            return "";
        }
        return null;
    }


    /**
     * Helper method to the category path construction operation.
     * Method parses name of the inner VCO category into the human-readable format.
     * @param cls VCO category 
     * @return string representation of the active VCO class
     */
    private String constructCategoryPath(OWLClass cls) {
        StringBuilder sb = new StringBuilder(cls.toStringID().split("VCO/")[1]);
        boolean again = true;
        while(again){
            Set<OWLClassExpression> express = cls.getSuperClasses(ontology);
            if(!express.isEmpty()){
                sb.append(";");
                for(OWLClassExpression cExp : express){
                    Set<OWLClass> classes = cExp.getClassesInSignature();
                    for(OWLClass owlClass : classes){
                        String owlStringId = owlClass.toStringID();
                        sb.append(owlStringId.split("VCO/")[1]);
                        cls = owlClass;
                    }
                }
            } else again = false;
        }
        String sbS = sb.toString();
        return sbS;
    }
        
    
    /**
     * Method returns the VCO category of the given synset. If the synset is 
     * not found in the VCO, it's hypernym (generalization) is used.
     * Method returns empty string if the ontology is not inicialized.
     * @param meaning synset to be assigned into a category
     * @param wnf Connection to the WordNet enabling WN operations
     * @return name of the category or empty string if the ontology is not inicialized.
     */
    public String getCategory(Synset meaning, WordNetOperations wnf){
        String category;
        if(meaning == null && !meaning.getPOS().equals(POS.NOUN)) return null;
        while(!meaning.equals(wnf.getPhysicalEntity()) && !meaning.equals(wnf.getAbstractEntity())){
            category = checkCategories(meaning);
            if(category == null) meaning = wnf.getHypernym(meaning);
            else if(category.length() == 0) return "";
            else return category;
        }
        return null;
    }
    
    /**
     * Method traverse the ontology top classes (top/"border" classes= classes dividing the 
     * space of the ontology into the particular logical areas) 
     * and extracts offsets of their synsets.
     *
     * Traversal of the ontology using external reasoner - PELLET REASONER 
     * see http://clarkparsia.com/pellet/
     * 
     * @return offsets of all synsets forming border classes of the VCO
     */
    public Set<Long> getLocatorsOfVcoBorderEntities(){
        Set<Long> locators = new HashSet<>(100);
        if(ontology != null){
            List<OWLClass> queue = new ArrayList<>();
            OWLOntologyManager manager = OWLManager.createOWLOntologyManager(); 
            // traverse in OWLOntology
            // get OWL_THING - root concept
            IRI classIRI = OWLRDFVocabulary.OWL_THING.getIRI();
            OWLClass rootClass = manager.getOWLDataFactory().getOWLClass(classIRI);
            // get subclasses.
            Set<OWLClass> rootSubClasses = reasoner.getSubClasses(rootClass, true).getFlattened();
            queue.addAll(rootSubClasses);
            while(!queue.isEmpty()){
                OWLClass active = queue.get(0);
                Set<OWLIndividual> individs = active.getIndividuals(ontology);
                if(individs.isEmpty()){
                    queue.addAll(reasoner.getSubClasses(active, true).getFlattened());
                } else{
                    for(OWLIndividual indiv : individs){
                        String offsS = indiv.toString().split("-")[3];
                        long offs = Long.parseLong(offsS);                        
                        offsetToClassNameMapping.put(offs, active);
                        locators.add(offs); 
                    }
                }
                queue.remove(0);
            }
        }
        return locators;
    }
    

    /**
     * Method returns list of names of all super classes for given synset.
     * Method returns null if the ontology is not inicialized.
     * @param active synset to be explored
     * @return list of all names of super classes or null if the ontology is not inicialized.
     */
    public List<String> getSuperCategories(Synset active) {
        if(ontology != null){
            List<String> IRIsOfSuperClass = new ArrayList<>();
            OWLClass activeOWLClass = offsetToClassNameMapping.get(active.getOffset());
            if(activeOWLClass == null) return null;
            Set<OWLClass> superClass = reasoner.getSuperClasses(activeOWLClass, false).getFlattened();
            for(OWLClass oClass : superClass){
                IRIsOfSuperClass.add(oClass.getIRI().toString());
            }
            return IRIsOfSuperClass;
        } 
        return null;
    }

    /**
     * Returns the level of the word (synset) in the VCO tree.
     * Method returns -1 if the ontology does not exist.
     * @param text
     * @return level of the word (synset) in the VCO tree
     */
    public int getLevel(String text) {    
        if(ontology == null){
            OWLOntologyManager manager = OWLManager.createOWLOntologyManager(); 
            IRI classIRI = OWLRDFVocabulary.OWL_THING.getIRI();
            OWLClass rootClass = manager.getOWLDataFactory().getOWLClass(classIRI);
            int searchedLevel = searchSubClasses(text, rootClass, 0);
            return searchedLevel;
        } return -1;
    }
    
    /**
     * It searches through the all VCO if the given vertexName (virtual vertex name)
     * is equal to the explored class name. 
     * It returns a depth on which the equal class was found.
     * 
     * @param vertexName
     * @param activeClass
     * @param level
     * @return level of the depth of the class in the VCO or -1 if the class is not 
     * presented in the VCO
     */
    private int searchSubClasses(String vertexName, OWLClass activeClass, int level){
        level++;
        for(OWLClass owlClass : reasoner.getSubClasses(activeClass, true).getFlattened()){
                String[] parsedIri = owlClass.getIRI().toString().split("/");
                String owlName = parsedIri[parsedIri.length-1];
                if(owlName.equals(vertexName)) return level;
                else{
                    int searchedLevel = searchSubClasses(vertexName, owlClass, level);
                    if(searchedLevel != -1) return searchedLevel;
                }
        }
        return -1;
    }

}
