
package messif.objects.annotation.util;

import messif.objects.annotation.SynsetRelationType;

/**
 * Class defines formulas to compute weight values of particular semantic
 * network edges.
 * For each type of relation in the system, individual formula exist.
 * @author Honza Botorek
 */
public class RelationsValueComputer {
    private final double hyponymConstant;
    private final double meronymConstant;
    private final double similarTermsConstant;
    private final double glossConstant;
    private final double lexFileConstant;

    public RelationsValueComputer(double hyponymConstant, double meronymConstant, double similarTermsConstant, double glossConstant, double lexFileConstant) {
        this.hyponymConstant = hyponymConstant;
        this.meronymConstant = meronymConstant;
        this.similarTermsConstant = similarTermsConstant;
        this.glossConstant = glossConstant;
        this.lexFileConstant = lexFileConstant;
    }

    
    
    //<editor-fold defaultstate="collapsed" desc="getters,setters">
    public  double getHyponymConstant() {
        return hyponymConstant;
    }

    public  double getMeronymConstant() {
        return meronymConstant;
    }

    public  double getSimilarTermsConstant() {
        return similarTermsConstant;
    }

    public  double getGlossConstant() {
        return glossConstant;
    }

    public  double getLexFileConstant() {
        return lexFileConstant;
    }
    
    
    
    //</editor-fold>
    
    /**
     * Factory method selects algorithm for the Relationship value computation
     * according to the RelationsEnum type.
     * @param relationsEnum type of the relationship/edge
     * @param membersCount amount of all vertices related to the active vertex by
     * given relationship
     * @return value of the probability distribution of the newly constructed edge
     */
    public double computeRelationshipValue(
            SynsetRelationType relationsEnum,
            int membersCount){
        double result = 0;
        switch(relationsEnum){
            case WN_HYPERNYM:
                result = computeHypernymRelation();
                break;
            case WN_HYPONYM:
                result = computeHyponymRelation(membersCount);
                break;
            case WN_MERONYM:
                result = computeMeronymRelation(membersCount);
                break;
            case WN_HOLONYM:
                result = computeHolonymRelation();
                break;
            case WN_SIMILAR_TERMS:
                result = computeSimilarTerms(membersCount);
                break;
            case WN_ADJ_SYNONYMS:
                result = computeAdjSynonym();
                break;
            case WN_GLOSS:
                result = computeGloss();
                break;
            case WN_VERB_LEX_FILE_NAME:
                result = computeVerbLexFile();
                break;
        }
        
        return result;
    }

    private double computeHypernymRelation() {
        return 1; //active.getVertexWord().getPoints();
    }

    private double computeHyponymRelation(int membersCount) {
        return (1 - hyponymConstant) / membersCount;
    }

    private double computeMeronymRelation(int membersCount) {
        return (1 - meronymConstant) / membersCount;
    }

    private double computeHolonymRelation() {
        return 1;
    }

    private double computeSimilarTerms(int membersCount) {
        return (1 - similarTermsConstant) / membersCount;
    }

    private double computeAdjSynonym() {
        return 1;
    }

    private double computeGloss() {
        return glossConstant;
    }

    private double computeVerbLexFile() {
        return lexFileConstant;
    }

}
