package messif.objects.annotation.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.data.Synset;

/**
 * Helper method that contains methods how to compute WN-based frequency. 
 * WordNet synsets are accompanied by a value that is extracted from corpus expressing
 * the frequency in texts. This measure may be utilized when initial vertex/term
 * probability is computed.
 * @author Honza Botorek
 */
public class WordNetFrequencyManager implements Comparator<Synset> {
    private final double maxFrequence;
    private final Map<POS, Map<Long, Integer>> synsetFreqMap;

    public WordNetFrequencyManager(File indexSenseFile) throws IOException {
        this.synsetFreqMap = new HashMap<>();
        for (Object pos : POS.getAllPOS()) {
            this.synsetFreqMap.put((POS)pos, new HashMap<Long, Integer>());
        }
        this.maxFrequence = initOffsetFreqMap(indexSenseFile, this.synsetFreqMap);
    }

    public int getSynsetFrequency(Synset synset) {
        return synsetFreqMap.get(synset.getPOS()).get(synset.getOffset());
    }

    public double getSynsetProbability(Synset synset) {
        Integer freq = getSynsetFrequency(synset);
        if (freq == null)
            return 0;
        return freq / maxFrequence;
    }

    @Override
    public int compare(Synset o1, Synset o2) {
        return Double.compare(getSynsetProbability(o2), getSynsetProbability(o1));
    }

    public List<Synset> getMostProbableSynsets(Collection<? extends Synset> synsets, int constWNMostProbabSynsets) {
        List<Synset> ret = new ArrayList<>(synsets);
        Collections.sort(ret, this);
        return ret.subList(0, Math.min(constWNMostProbabSynsets, ret.size()));
    }

    public Map<Synset, Integer> getSynsetFrequencies(Collection<? extends Synset> synsets) {
        Map<Synset, Integer> ret = new HashMap<>();
        for (Synset synset : synsets) {
            ret.put(synset, getSynsetFrequency(synset));
        }
        return ret;
    }

    private static POS getSenseTypePOS(char senseType) {
        switch (senseType) {
            case '1':
                return POS.NOUN;
            case '2':
                return POS.VERB;
            case '3':
                return POS.ADJECTIVE;
            case '4':
                return POS.ADVERB;
            default:
                return POS.ADJECTIVE;
        }
    }

    /**
     * Method loads frequency file into the memory
     */
    private static int initOffsetFreqMap(File indexSenseFile, Map<POS, Map<Long, Integer>> synsetFreqMap) throws IOException {
        int maxFrequence = 0;
        BufferedReader freqReader = new BufferedReader(new FileReader(indexSenseFile));
        String line = freqReader.readLine();
        while(line != null){
            String[] lineSplitted = line.split("\\s");
            Map<Long, Integer> synsetOffsetMap = synsetFreqMap.get(getSenseTypePOS(lineSplitted[0].charAt(lineSplitted[0].indexOf('%') + 1)));
            long senseLongId = Long.parseLong(lineSplitted[1]);
            int frequence = Integer.parseInt(lineSplitted[3]);
            if (maxFrequence < frequence)
                maxFrequence = frequence;
            Integer currFreq = synsetOffsetMap.get(senseLongId);
            if (currFreq == null || currFreq < frequence) {
                synsetOffsetMap.put(senseLongId, frequence);
            }
            line = freqReader.readLine();
        }
        freqReader.close();
        return maxFrequence;
    }

    //////////////////////// OLD COMPUTATION (HACKS BELOW THIS LINE!!!!) ////////////////////////

    /**
     * Method selects certain amount of the synsets according to the given parameter.
     * @param synsets all synsets to consider for the most frequent
     * @param wnMostProbabSyns number defining the amount of returning synsets
     * @return certain amount of the synsets according to the given parameter.
     */
    public Map<Synset, Double> selectMostFrequentNormalized(Collection<? extends Synset> synsets, int wnMostProbabSyns) {
        return normalizeMostFrequentSyns(selectMostFrequent(getSynsetFrequencies(synsets), wnMostProbabSyns));
    }

    private static Map<Synset, Integer> selectMostFrequent(Map<Synset, Integer> synsetMap, int wnMostProbabSyns) {
        Map<Synset, Integer> accepted = new HashMap<>();
        for(Synset synI : synsetMap.keySet()){
            int maxFreq = -1;
            Synset maxSyn = null;
            for(Synset synJ : synsetMap.keySet()){
                int currFreq = synsetMap.get(synJ);
                if(currFreq > maxFreq
                        && !accepted.keySet().contains(synJ)){
                    maxFreq = currFreq;
                    maxSyn = synJ;
                }
            }
            accepted.put(maxSyn, maxFreq + 1);
            wnMostProbabSyns--;
            if(wnMostProbabSyns == 0) return accepted;
        }
        return accepted;
    }

    private static Map<Synset, Double> normalizeMostFrequentSyns(Map<Synset, Integer> mostFrequentSyns) {
        Map<Synset, Double> normalizedMostFrequentSyns = new HashMap<>();
        double sum = 0;
        for(Synset syn : mostFrequentSyns.keySet()){
            sum += mostFrequentSyns.get(syn);
        }
        for(Synset syn : mostFrequentSyns.keySet()){
            double originalValue = mostFrequentSyns.get(syn);
            double normalizedValue = originalValue / sum;
            normalizedMostFrequentSyns.put(syn, normalizedValue);
        }
        return normalizedMostFrequentSyns;
    }
}
