package messif.objects.annotation.util;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathFactory;
import net.didion.jwnl.JWNL;
import net.didion.jwnl.JWNLException;
import net.didion.jwnl.data.IndexWord;
import net.didion.jwnl.data.IndexWordSet;
import net.didion.jwnl.data.POS;
import net.didion.jwnl.data.Pointer;
import net.didion.jwnl.data.PointerType;
import net.didion.jwnl.data.Synset;
import net.didion.jwnl.data.Word;
import net.didion.jwnl.data.list.PointerTargetNode;
import net.didion.jwnl.data.list.PointerTargetNodeList;
import net.didion.jwnl.data.relationship.Relationship;
import net.didion.jwnl.data.relationship.RelationshipFinder;
import net.didion.jwnl.data.relationship.RelationshipList;
import net.didion.jwnl.dictionary.Dictionary;
import org.w3c.dom.Document;

/**
 * Functions that extend WordNetOperationsInterface. This class provide higher
 * functions with WordNet relations. It leverages underlying basic
 * WordNetOperations.
 *
 * @author Honza Botorek
 */
public class WordNetOperations {

    private final String dictionaryPath;
    private final Dictionary dictionary;
    private final Synset PHYSICAL_OBJECT;
    //private final Synset THING;
    private final Synset PHYSICAL_ENTITY;
    private final Synset ABSTRACT_ENTITY;
    private final Synset ENTITY;

    public WordNetOperations(String confPath) {
        try {
            try (FileInputStream in = new FileInputStream(confPath)) {
                JWNL.initialize(in);
            }
            try (FileInputStream in = new FileInputStream(confPath)) {
                XPathExpression xpath = XPathFactory.newInstance().newXPath().compile("//param[@name = 'dictionary_path']/@value");
                Document doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().parse(in);
                dictionaryPath = xpath.evaluate(doc);
            }
            dictionary = Dictionary.getInstance();
        } catch (Exception ex) {
            throw new IllegalArgumentException("Cannot initialize WordNet: " + ex, ex);
        }

        // Initialize constants
        PHYSICAL_OBJECT = getSynsets(POS.NOUN, "object")[0];
        PHYSICAL_ENTITY = getSynsets(POS.NOUN, "physical entity")[0];
        ABSTRACT_ENTITY = getSynsets(POS.NOUN, "abstract entity")[0];
        ENTITY = getSynsets(POS.NOUN, "entity")[0];
    }

    public String getDictionaryPath() {
        return dictionaryPath;
    }

    public boolean isPhysicalObject(Synset synset) {
        try {
            RelationshipList list = RelationshipFinder.getInstance().findRelationships(synset, PHYSICAL_OBJECT, PointerType.HYPERNYM);
            if (!list.isEmpty()) {
                Relationship rel = (Relationship)list.get(0);
                PointerTargetNodeList ptnl = rel.getNodeList();                         // transform it to the list of pointers
                for (int i = 0; i < ptnl.size(); i++) {
                    PointerTargetNode pointerTargetNode = (PointerTargetNode)ptnl.get(i);
                    if (pointerTargetNode.getSynset().equals(ENTITY)
                            || pointerTargetNode.getSynset().equals(PHYSICAL_ENTITY)) {
                        return false;
                    }
                }
                return true;
            } else
                return false;
        } catch (JWNLException ex) {
            throw new IllegalStateException("countPath error : " + ex, ex);
        }
    }

    /**
     * Method returns direct hypernym of given NOUN synset. Null if the synset
     * is not a NOUN.
     *
     * @param syn
     * @return
     */
    public Synset getDirectHypernym(Synset syn) {
        if (syn.getPOS().equals(POS.NOUN))
            return getHypernym(syn);
        return null;
    }
    
    /**
     * Method returns direct hypernym of given VERB synset. Null if the synset
     * is not a NOUN.
     *
     * @param syn
     * @return
     */
    public Synset getDirectVerbHypernym(Synset syn) {
        if (syn.getPOS().equals(POS.VERB))
            return getHypernym(syn);
        return null;
    }

    /**
     * private method used for obtaining pointers from given synset. It returns
     * list of synsets reachable by specified pointer type from input synset. If
     * such pointer doesn't exist - the method returns empty list.
     *
     * @param syn
     * @param point
     * @return
     */
    private List<Synset> getPointers(Synset syn, PointerType point) {
        List<Synset> pointedSynsets = new ArrayList<Synset>();
        Pointer[] pointers = syn.getPointers(point);
        for (Pointer p : pointers) {
            try {
                pointedSynsets.add(p.getTargetSynset());
            } catch (JWNLException ex) {
                throw new IllegalStateException(ex);
            }
        }
        return pointedSynsets;
    }
    
    public List<Synset> getAdjSynonyms(Synset syn){
        return getPointers(syn, PointerType.SIMILAR_TO);
    }
    

    /**
     * Simple method using getPointers method. It returns domain terms for given
     * synset.
     *
     * @param syn
     * @return
     */
    public List<Synset> getDomainTerms(Synset syn) {
        return getPointers(syn, PointerType.CATEGORY_MEMBER);
    }

    /**
     * Simple method using getPointers method. It returns nominalization terms
     * for given synset.
     *
     * @param syn
     * @return
     */
    public List<Synset> getNominalizationTerms(Synset syn) {
        return getPointers(syn, PointerType.NOMINALIZATION);
    }

    /**
     * Simple method using getPointers method. It returns direct meronyms terms
     * for given synset.
     *
     * @param syn
     * @return
     */
    public List<Synset> getDirectMeronyms(Synset syn) {
        return getPointers(syn, PointerType.PART_MERONYM);
    }
    
    /**
     * Simple method using getPointers method. It returns direct holonym terms
     * for given synset.
     *
     * @param syn
     * @return
     */
    public List<Synset> getDirectHolonyms(Synset syn){
        return getPointers(syn, PointerType.PART_HOLONYM);
    }
    
    /**
     * Simple method using getPointers method. It returns direct hyponym terms
     * for given synset.
     *
     * @param syn
     * @return
     */
    public List<Synset> getDirectHyponyms(Synset syn){
        return getPointers(syn, PointerType.HYPONYM);
    }

    /**
     * Method returns meronyms for all synsets from the hypernymy path from
     * given synset up to physical object(most generic one).
     *
     * @param syn
     * @return
     */
    public List<Synset> getAllMeronyms(Synset syn) {
        List<Synset> output = new ArrayList<Synset>();
        Synset currentSyn = syn;
        if (currentSyn.getPOS().equals(POS.NOUN)) {
            while (!currentSyn.equals(ENTITY)) {
                output.addAll(getDirectMeronyms(currentSyn));
                currentSyn = getDirectHypernym(currentSyn);
            }
        }
        return output;
    }


    /**
     * Examines, if the given string (built from 2 independent words) has some
     * mapping to the synset in WN database.
     *
     * @param possibleWord
     * @return
     */
    public Synset[] getNounSynsetFromGroupedWords(String possibleWord) {
        return getGroupedWordSynsets(possibleWord);

    }

    public Synset getSynsetAt(long offset, POS pos) {
        return getSynsetById(offset, pos);
    }

    /**
     * Method returns gloss for given synset.
     *
     * @param syn
     * @return
     */
    public String getGloss(Synset syn) {
        return syn.getGloss();
    }

    /**
     * Returns array of synsets for given string and word position.
     *
     * @param pos
     * @param word
     * @return
     */
    public Synset[] getSynsets(POS pos, String word) {
        try {
            IndexWord iw = dictionary.lookupIndexWord(pos, word);
            if (iw != null) {
                return iw.getSenses();
            } else
                return null;
        } catch (java.lang.Exception ex) {
            throw new IllegalStateException("getSynsets error : " + ex, ex);
        }
    }
    
    public List<Synset> getAllSynsets(String word){
        List<Synset> all = new ArrayList<>();
        Synset[] nouns = getSynsets(POS.NOUN, word);
        Synset[] adjectives = getSynsets(POS.ADJECTIVE, word);
        Synset[] adverbs = getSynsets(POS.ADVERB, word);
        Synset[] verbs = getSynsets(POS.VERB, word);
        if(nouns!=null) all.addAll(Arrays.asList(nouns));
        if(verbs!=null) all.addAll(Arrays.asList(verbs));
        if(adjectives!=null) all.addAll(Arrays.asList(adjectives));
        if(adverbs!=null) all.addAll(Arrays.asList(adverbs));
        return all;
    }

    /**
     * It returns direct hypernym for given NOUN synset; null if there is no
     * hypernym for given synset.
     *
     * @param syn
     * @return
     */
    public Synset getHypernym(Synset syn) {
        try {
            Pointer[] hypernym = syn.getPointers(PointerType.HYPERNYM);
            if (hypernym.length != 0)
                return hypernym[0].getTargetSynset();
            return null;
        } catch (JWNLException ex) {
            throw new IllegalStateException("getHypernym error : " + ex, ex);
        }
    }

    /**
     * It returns array of all direct hyponyms for given synset; null if there
     * are no hyponyms.
     *
     * @param syn
     * @return
     */
    public Synset[] getHyponyms(Synset syn) {
        try {
            Pointer[] hyponymsPointer = syn.getPointers(PointerType.HYPONYM);
            Synset[] hyponyms = new Synset[hyponymsPointer.length];
            for (int i = 0; i < hyponymsPointer.length; i++) {
                hyponyms[i] = hyponymsPointer[i].getTargetSynset();
            }
            return hyponyms;
        } catch (JWNLException ex) {
            throw new IllegalStateException("getHyponym err: " + ex, ex);
        }
    }

    /**
     * Simple method that decides whether or not the given word can be found in
     * the WN.
     *
     * @param word
     * @return
     */
    public boolean isInWn(String word) {
        try {
            IndexWordSet iws = dictionary.lookupAllIndexWords(word);
            return iws.size() != 0;
        } catch (Exception ex) {
            throw new IllegalStateException(ex);
        }
    }

    /**
     * Method decides if the word is a NOUN.
     *
     * @param word
     * @return
     */
    public boolean isNoun(String word) {
        try {
            IndexWordSet iws = dictionary.lookupAllIndexWords(word);
            for (IndexWord iw : iws.getIndexWordArray()) {
                if (iw.getPOS().equals(POS.NOUN))
                    return true;
            }
            return false;
        } catch (JWNLException ex) {
            throw new IllegalStateException(ex);
        }
    }

    /**
     * Method decides if the word is an instance.
     *
     * @param word
     * @return
     */
    public boolean isInstance(Synset syn) {
        Pointer[] temp = syn.getPointers(PointerType.INSTANCE_HYPERNYM);
        return temp.length != 0;
    }

    /**
     * Method returns direct meronyms of given synset. If the synset doesn't
     * have any meronym pointers - it returns empty array.
     *
     * @param syn
     * @return
     */
    public Synset[] getMeronyms(Synset syn) {
        try {
            Pointer[] meronymsPointer = syn.getPointers(PointerType.PART_MERONYM);
            Synset[] meronyms = new Synset[meronymsPointer.length];
            for (int i = 0; i < meronymsPointer.length; i++) {
                meronyms[i] = meronymsPointer[i].getTargetSynset();
            }
            return meronyms;
        } catch (JWNLException ex) {
            throw new IllegalStateException("getMeronyms err: " + ex, ex);
        }
    }

    /**
     * Simple method that retrieve synset from database by given offset.
     *
     * @param word
     * @return
     */
    public Synset getSynsetById(long id, POS pos) {
        try {
            return dictionary.getSynsetAt(pos, id);
        } catch (Exception ex) {
            throw new IllegalStateException(ex);
        }
    }

    /**
     * Method used for seraching of grouped words. It recieves a two-word
     * expression as a parameter and decides whether or not there is such synset
     * in the WN. It returns such synset; null otherwise.
     *
     * @param word
     * @return
     */
    public Synset[] getGroupedWordSynsets(String word) {
        try {
            IndexWord iw = dictionary.getIndexWord(POS.NOUN, word);
            if (iw != null)
                return iw.getSenses();
            else
                return null;
        } catch (JWNLException ex) {
            throw new IllegalStateException(ex);
        }
    }

    /**
     * Simple method searching across all positions of WN and trying to find all
     * possible meanings for given word. If there is only one UNIQUE meaning -
     * it returns such synset. Otherwise it returns null;
     *
     * @param word
     * @return
     */
    public Synset isUnique(String word) {
        List<Synset> temp = new ArrayList<Synset>();
        for (Object p : POS.getAllPOS()) {
            Synset[] tempMeanings = getSynsets((POS)p, word);
            if (tempMeanings != null)
                temp.addAll(Arrays.asList(tempMeanings));
        }
        if (temp.size() == 1)
            return temp.get(0);
        else
            return null;
    }

    /**
     * Simple method searching unique meaning in NOUN category of given string.
     * If there is only one UNIQUE meaning - it returns such synset. Otherwise
     * it returns null;
     *
     * @param word
     * @return
     */
    public Synset isUniqueNoun(String word) {
        Synset[] tempMeanings = getSynsets(POS.NOUN, word);
        if (tempMeanings != null && tempMeanings.length == 1) {
            return tempMeanings[0];
        } else
            return null;
    }

    /**
     * Method that returns direct synonyms - all lemmas from synset's "name"
     *
     * @param syn
     * @return
     */
    public Set<String> getLemmasFromSynset(Synset syn) {
        if (syn == null)
            return null;
        Set<String> lemmas = new HashSet<String>();
        Word[] words = syn.getWords();
        for (Word w : words) {
            lemmas.add(w.getLemma());
        }
        return lemmas;
    }
    
    /**
     * 
     * @return Map<Synset,Long>
     */
    public Map<Synset,Long> getNumberOfAllHyponymsForVitalSynsets(){
        Map<Synset,Long> hypoCount = new HashMap<>(90000);
        try {
            BufferedReader reader = new BufferedReader(new InputStreamReader(getClass().getClassLoader().getResourceAsStream("resources/HyponymyCount.csv")));
            String line = reader.readLine();
            while(line != null){
                String[] splitted = line.split(":");
                long id = Long.parseLong(splitted[0]);
                Synset tempSynset = getSynsetById(id, POS.NOUN);
                long hypoCountVal = Long.parseLong(splitted[1]);
                hypoCount.put(tempSynset, hypoCountVal);
                line = reader.readLine();
            }
        } catch (Exception ex) {
            throw new IllegalStateException("HyponymCount file was not found.", ex);
        }
        return hypoCount;
    }
    
    public long getAmountOfAllHyponyms(Synset syn){
        Pointer[] points = syn.getPointers(PointerType.HYPONYM);
        long hypocount = 0;
        for(Pointer p : points){
            try {
                hypocount += getHypoCount(p.getTargetSynset(), hypocount);
            } catch (JWNLException ex) {
                throw new IllegalStateException(ex);
            }
        }        
        return hypocount;
    }
    
    /**
     * Method returns number of hyponyms for given synset. 
     * It represents recursive function - long hypocount is given as an attribute
     * that counts each particular hyponym. If currently processed word has also hyponyms,
     * method is recursively called on offsprings.
     * @param syn synset that is checked for all hyponyms.
     * @param hypocount counter summing the overall hyponym amount.
     * @return number of all hyponyms for given synset
     * @throws JWNLException when 
     */
    private long getHypoCount(Synset syn, long hypocount){
        long tempHypoCount = hypocount;
        hypocount++;
        Pointer[] points = syn.getPointers(PointerType.HYPONYM);
        if(points.length == 0){
            return hypocount-tempHypoCount;
        } else{
            for(Pointer p : points){
                try {
                    hypocount += getHypoCount(p.getTargetSynset(),hypocount);
                } catch (JWNLException ex) {
                    throw new IllegalStateException(ex);
                }
            }
            return hypocount-tempHypoCount;
        }
    }
    
    /**
     * Method returns the top of Synset NOUNs tree - the ENTITY synset.
     * @return 
     */
    public Synset getCentralBorderEntity(){
        return ENTITY;
    }

    /**
     * Method returns the top of physical entity subtree - the Physical entity synset
     * @return 
     */
    public Synset getPhysicalEntity(){
        return PHYSICAL_ENTITY;
    }
    
    /**
     * Method returns the top of abstract entity subtree - the Abstract entity synset
     * @return 
     */
    public Synset getAbstractEntity(){
        return ABSTRACT_ENTITY;
    }

    public Dictionary getDictionary() {
        return dictionary;
    }
}

